"""
   Job Base Class

   This class provides generic job definition functionality suitable for any VO.

   Helper functions are documented with example usage for the DIRAC API.  An example
   script (for a simple executable) would be::

     from DIRAC.Interfaces.API.Dirac import Dirac
     from DIRAC.Interfaces.API.Job import Job

     j = Job()
     j.setCPUTime(500)
     j.setExecutable('/bin/echo hello')
     j.setExecutable('yourPythonScript.py')
     j.setExecutable('/bin/echo hello again')
     j.setName('MyJobName')

     dirac = Dirac()
     jobID = dirac.submitJob(j)
     print 'Submission Result: ',jobID

   Note that several executables can be provided and wil be executed sequentially.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import re
import os
import shlex

import six
from six import StringIO
from six.moves.urllib.parse import quote as urlquote

from DIRAC import S_OK, gLogger
from DIRAC.Core.Base.API import API
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Core.Workflow.Parameter import Parameter
from DIRAC.Core.Workflow.Workflow import Workflow
from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd
from DIRAC.Core.Utilities.Subprocess import systemCall
from DIRAC.Core.Utilities.List import uniqueElements
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOForGroup
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getCESiteMapping
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.Workflow.Utilities.Utils import getStepDefinition, addStepToWorkflow


COMPONENT_NAME = "/Interfaces/API/Job"


class Job(API):
    """DIRAC jobs"""

    #############################################################################

    def __init__(self, script=None, stdout="std.out", stderr="std.err"):
        """Instantiates the Workflow object and some default parameters."""

        super(Job, self).__init__()

        self.stepCount = 0
        self.owner = "NotSpecified"
        self.name = "Name"
        self.type = "User"
        self.priority = 1
        vo = ""
        ret = getProxyInfo(disableVOMS=True)
        if ret["OK"] and "group" in ret["Value"]:
            vo = getVOForGroup(ret["Value"]["group"])
        self.group = vo
        self.site = None
        self.stdout = stdout
        self.stderr = stderr
        self.logLevel = "INFO"
        self.executable = "dirac-jobexec"  # to be clarified
        self.addToInputSandbox = []
        self.addToOutputSandbox = []
        self.addToInputData = []
        # #Add member to handle Parametric jobs
        self.numberOfParameters = 0
        self.parameterSeqs = {}
        self.wfArguments = {}
        self.parametricWFArguments = {}

        # loading the function that will be used to determine the platform (it can be VO specific)
        res = ObjectLoader().loadObject("ConfigurationSystem.Client.Helpers.Resources", "getDIRACPlatforms")
        if not res["OK"]:
            self.log.fatal(res["Message"])
        self.getDIRACPlatforms = res["Value"]

        self.script = script
        if not script:
            self.workflow = Workflow()
            self.__setJobDefaults()
        else:
            self.workflow = Workflow(script)

    #############################################################################

    def setExecutable(self, executable, arguments="", logFile="", modulesList=None, parameters=None, paramValues=None):
        """Helper function.

        Specify executable script to run with optional arguments and log file
        for standard output.

        These can be either:

        - Submission of a python or shell script to DIRAC
           - Can be inline scripts e.g. C{'/bin/ls'}
           - Scripts as executables e.g. python or shell script file

        Example usage:

        >>> job = Job()
        >>> job.setExecutable('myScript.py')

        :param str executable: Executable
        :param str arguments: Optional arguments to executable
        :param str logFile: Optional log file name
        :param list modulesList: Optional list of modules (to be used mostly when extending this method)
        :param parameters: Optional list of parameters (to be used mostly when extending this method)
        :type parameters: python:list of tuples
        :param paramValues: Optional list of parameters values (to be used mostly when extending this method)
        :type parameters: python:list of tuples
        """
        kwargs = {"executable": executable, "arguments": arguments, "logFile": logFile}
        if (
            not isinstance(executable, six.string_types)
            or not isinstance(arguments, six.string_types)
            or not isinstance(logFile, six.string_types)
        ):
            return self._reportError("Expected strings for executable and arguments", **kwargs)

        if os.path.exists(executable):
            self.log.verbose("Found script executable file %s" % (executable))
            self.addToInputSandbox.append(executable)
            logName = "%s.log" % (os.path.basename(executable))
        else:
            self.log.warn("The executable code could not be found locally")
            logName = "CodeOutput.log"

        self.stepCount += 1
        stepName = "RunScriptStep%s" % (self.stepCount)

        if logFile:
            if isinstance(logFile, six.string_types):
                logName = str(logFile)
        else:
            logName = "Script%s_%s" % (self.stepCount, logName)

        if not modulesList:
            modulesList = ["Script"]
        if not parameters:
            parameters = [
                ("executable", "string", "", "Executable Script"),
                ("arguments", "string", "", "Arguments for executable Script"),
                ("applicationLog", "string", "", "Log file name"),
            ]

        step = getStepDefinition("ScriptStep%s" % (self.stepCount), modulesList, parametersList=parameters)
        self.addToOutputSandbox.append(logName)

        stepInstance = addStepToWorkflow(self.workflow, step, stepName)

        stepInstance.setValue("applicationLog", logName)
        stepInstance.setValue("executable", executable)
        if arguments:
            # If arguments are expressed in terms of parameters, pass them to Workflow
            # These arguments will be resolved on the server side for each parametric job
            if re.search(r"%\(.*\)s", arguments) or re.search("%n", arguments):
                self.parametricWFArguments["arguments"] = arguments
            else:
                stepInstance.setValue("arguments", arguments)
        if paramValues:
            for param, value in paramValues:
                stepInstance.setValue(param, value)

        return S_OK(stepInstance)

    #############################################################################
    def setName(self, jobName):
        """Helper function.

        A name for the job can be specified if desired. This will appear
        in the JobName field of the monitoring webpage. If nothing is
        specified a default value will appear.

        Example usage:

        >>> job=Job()
        >>> job.setName("myJobName")

        :param str jobName: Name of job
        """
        kwargs = {"jobname": jobName}
        if not isinstance(jobName, six.string_types):
            return self._reportError("Expected strings for job name", **kwargs)

        self.workflow.setName(jobName)
        self._addParameter(self.workflow, "JobName", "JDL", jobName, "User specified name")

        return S_OK()

    #############################################################################
    def setInputSandbox(self, files):
        """Helper function.

        Specify input sandbox files less than 10MB in size.  If over 10MB, files
        or a directory may be uploaded to Grid storage, see C{dirac.uploadSandbox()}.

        Paths to the options file and (if required) 'lib/' directory of the DLLs
        are specified here. Default is local directory.
        Executables may be placed in the lib/ directory if desired. The lib/ directory
        is transferred to the Grid Worker Node before the job executes.

        Files / directories can be specified using the `*` character e.g. `*.txt`  these
        are resolved correctly before job execution on the WN.

        Example usage:

        >>> job = Job()
        >>> job.setInputSandbox(['DaVinci.opts'])

        :param files: Input sandbox files, can specify full path
        :type files: Single string or list of strings ['','']
        """
        if isinstance(files, list) and files:
            resolvedFiles = self._resolveInputSandbox(files)
            fileList = ";".join(resolvedFiles)
            description = "Input sandbox file list"
            self._addParameter(self.workflow, "InputSandbox", "JDL", fileList, description)
            # self.sandboxFiles=resolvedFiles
        elif isinstance(files, six.string_types):
            resolvedFiles = self._resolveInputSandbox([files])
            fileList = ";".join(resolvedFiles)
            description = "Input sandbox file"
            # self.sandboxFiles = [files]
            self._addParameter(self.workflow, "InputSandbox", "JDL", fileList, description)
        else:
            kwargs = {"files": files}
            return self._reportError("Expected file string or list of files for input sandbox contents", **kwargs)

        return S_OK()

    #############################################################################

    def setOutputSandbox(self, files):
        """Helper function.

        Specify output sandbox files. If specified files are over 10MB, these
        may be uploaded to Grid storage with a notification returned in the
        output sandbox.

        Example usage:

        >>> job = Job()
        >>> job.setOutputSandbox(['DaVinci_v19r12.log','DVNTuples.root'])

        :param files: Output sandbox files
        :type files: Single string or list of strings ['','']

        """
        if isinstance(files, list) and files:
            fileList = ";".join(files)
            description = "Output sandbox file list"
            self._addParameter(self.workflow, "OutputSandbox", "JDL", fileList, description)
        elif isinstance(files, six.string_types):
            description = "Output sandbox file"
            self._addParameter(self.workflow, "OutputSandbox", "JDL", files, description)
        else:
            kwargs = {"files": files}
            return self._reportError("Expected file string or list of files for output sandbox contents", **kwargs)

        return S_OK()

    #############################################################################

    def setInputData(self, lfns):
        """Helper function.

        Specify input data by Logical File Name (LFN).

        Example usage:

        >>> job = Job()
        >>> job.setInputData(['/lhcb/production/DC04/v2/DST/00000742_00003493_10.dst'])

        :param lfns: Logical File Names
        :type lfns: Single LFN string or list of LFNs
        """
        if isinstance(lfns, list) and lfns:
            for i, _ in enumerate(lfns):
                lfns[i] = lfns[i].replace("LFN:", "")
            inputData = ["LFN:" + x for x in lfns]
            inputDataStr = ";".join(inputData)
            description = "List of input data specified by LFNs"
            self._addParameter(self.workflow, "InputData", "JDL", inputDataStr, description)
        elif isinstance(lfns, six.string_types):  # single LFN
            description = "Input data specified by LFN"
            self._addParameter(self.workflow, "InputData", "JDL", lfns, description)
        else:
            kwargs = {"lfns": lfns}
            return self._reportError("Expected lfn string or list of lfns for input data", **kwargs)

        return S_OK()

    #############################################################################

    def setParameterSequence(self, name, parameterList, addToWorkflow=False):
        """Function to define a sequence of values for parametric jobs.

        :param str name: sequence parameter name
        :param list parameterList: list of parameter values
        :param bool addToWorkflow: flag to add parameter to the workflow on the fly, if str, then
                                   use as the workflow parameter
        :return: S_OK/S_ERROR
        """

        if self.numberOfParameters == 0:
            self.numberOfParameters = len(parameterList)
        elif self.numberOfParameters != len(parameterList):
            return self._reportError("Parameter sequences of different length", name="setParameterSequence")

        self.parameterSeqs[name] = parameterList
        if addToWorkflow:
            if isinstance(addToWorkflow, six.string_types):
                self.wfArguments[name] = addToWorkflow
            else:
                self.wfArguments[name] = name

        return S_OK()

    #############################################################################
    def setInputDataPolicy(self, policy, dataScheduling=True):
        """Helper function.

        Specify a job input data policy, this takes precedence over any site specific or
        global settings.

        Possible values for policy are 'Download' or 'Protocol' (case-insensitive). This
        requires that the module locations are defined for the VO in the CS.

        Example usage:

        >>> job = Job()
        >>> job.setInputDataPolicy('download')

        """
        kwargs = {"policy": policy, "dataScheduling": dataScheduling}
        csSection = "InputDataPolicy"
        possible = ["Download", "Protocol"]
        finalPolicy = ""
        for value in possible:
            if policy.lower() == value.lower():
                finalPolicy = value

        if not finalPolicy:
            return self._reportError(
                "Expected one of %s for input data policy" % (", ".join(possible)), __name__, **kwargs
            )

        jobPolicy = Operations().getValue("%s/%s" % (csSection, finalPolicy), "")
        if not jobPolicy:
            return self._reportError(
                "Could not get value for Operations option %s/%s" % (csSection, finalPolicy), __name__, **kwargs
            )

        description = "User specified input data policy"
        self._addParameter(self.workflow, "InputDataPolicy", "JDL", jobPolicy, description)

        if not dataScheduling and policy.lower() == "download":
            self.log.verbose("Scheduling by input data is disabled, jobs will run anywhere and download input data")
            self._addParameter(
                self.workflow, "DisableDataScheduling", "JDL", "True", "Disable scheduling by input data"
            )

        if not dataScheduling and policy.lower() != "download":
            self.log.error('Expected policy to be "download" for bypassing data scheduling')
            return self._reportError(
                'Expected policy to be "download" for bypassing data scheduling', __name__, **kwargs
            )

        return S_OK()

    #############################################################################
    def setOutputData(self, lfns, outputSE=None, outputPath=""):
        """Helper function.

        For specifying output data to be registered in Grid storage.  If a list
        of OutputSEs are specified the job wrapper will try each in turn until
        successful.  If the OutputPath is specified this will appear only after
        / <VO> / user / <initial> / <username>
        directory.

        The output data can be LFNs or local file names.
        If they are LFNs they should be pre-prended by "LFN:",
        otherwise they will be interpreted as local files to be found.
        If local files are specified, then specifying the outputPath may become necessary,
        because if it's not specified then it will be constructed starting from the user name.

        Example usage:

        >>> job = Job()
        >>> job.setOutputData(['DVNtuple.root'])

        :param lfns: Output data file or files.
        :type lfns: Single string or list of strings ['','']
        :param outputSE: Optional parameter to specify the Storage Element to store data or files, e.g. CERN-tape
        :type outputSE: string or python:list
        :param outputPath: Optional parameter to specify part of the path in the storage (see above)
        :type outputPath: string

        """
        if outputSE is None:
            outputSE = []
        kwargs = {"lfns": lfns, "OutputSE": outputSE, "OutputPath": outputPath}
        if isinstance(lfns, list) and lfns:
            outputDataStr = ";".join(lfns)
            description = "List of output data files"
            self._addParameter(self.workflow, "OutputData", "JDL", outputDataStr, description)
        elif isinstance(lfns, six.string_types):
            description = "Output data file"
            self._addParameter(self.workflow, "OutputData", "JDL", lfns, description)
        else:
            return self._reportError("Expected file name string or list of file names for output data", **kwargs)

        if outputSE:
            description = "User specified Output SE"
            if isinstance(outputSE, six.string_types):
                outputSE = [outputSE]
            elif not isinstance(outputSE, list):
                return self._reportError("Expected string or list for OutputSE", **kwargs)
            outputSE = ";".join(oSE.strip() for oSE in outputSE)
            self._addParameter(self.workflow, "OutputSE", "JDL", outputSE, description)

        if outputPath:
            description = "User specified Output Path"
            if not isinstance(outputPath, six.string_types):
                return self._reportError("Expected string for OutputPath", **kwargs)
            # Remove leading "/" that might cause problems with os.path.join
            # This will prevent to set OutputPath outside the Home of the User
            while outputPath[0] == "/":
                outputPath = outputPath[1:]
            self._addParameter(self.workflow, "OutputPath", "JDL", outputPath, description)

        return S_OK()

    #############################################################################
    def setPlatform(self, platform):
        """Developer function: sets the target platform, e.g. Linux_x86_64_glibc-2.17.
        This platform is in the form of what it is returned by the dirac-platform script
        (or dirac-architecture if your extension provides it)
        """
        kwargs = {"platform": platform}

        if not isinstance(platform, six.string_types):
            return self._reportError("Expected string for platform", **kwargs)

        if not platform.lower() == "any":
            availablePlatforms = self.getDIRACPlatforms()
            if not availablePlatforms["OK"]:
                return self._reportError("Can't check for platform", **kwargs)
            if platform in availablePlatforms["Value"]:
                self._addParameter(self.workflow, "Platform", "JDL", platform, "Platform ( Operating System )")
            else:
                return self._reportError("Invalid platform", **kwargs)

        return S_OK()

    #############################################################################
    def setSubmitPool(self, backend):
        """Developer function.

        Choose submission pool on which job is executed.
        Default in place for users.
        """
        # should add protection here for list of supported platforms
        kwargs = {"backend": backend}
        if not isinstance(backend, six.string_types):
            return self._reportError("Expected string for SubmitPool", **kwargs)

        if not backend.lower() == "any":
            self._addParameter(self.workflow, "SubmitPools", "JDL", backend, "Submit Pool")

        return S_OK()

    #############################################################################
    def setCPUTime(self, timeInSecs):
        """Helper function.

        Example usage:

        >>> job = Job()
        >>> job.setCPUTime(5000)

        :param timeInSecs: CPU time
        :type timeInSecs: int
        """
        kwargs = {"timeInSecs": timeInSecs}
        if not isinstance(timeInSecs, int):
            try:
                timeInSecs = int(timeInSecs)
            except ValueError:
                if not re.search("{{", timeInSecs):
                    return self._reportError("Expected numerical string or int for CPU time in seconds", **kwargs)

        description = "CPU time in secs"
        self._addParameter(self.workflow, "CPUTime", "JDL", timeInSecs, description)
        return S_OK()

    #############################################################################
    def setDestination(self, destination):
        """Helper function.

        Can specify a desired destination site or sites for job.  This can be useful
        for debugging purposes but often limits the possible candidate sites
        and overall system response time.

        Example usage:

        >>> job = Job()
        >>> job.setDestination('LCG.CERN.ch')

        :param destination: site string
        :type destination: str or python:list

        :return: S_OK/S_ERROR
        """
        kwargs = {"destination": destination}
        if isinstance(destination, six.string_types):
            destination = destination.replace(" ", "").split(",")
            description = "User specified destination site"
        else:
            description = "List of sites selected by user"
        if isinstance(destination, list):
            sites = set(site for site in destination if site.lower() != "any")
            if sites:
                result = self._checkSiteIsValid(sites)
                if not result["OK"]:
                    return self._reportError("%s is not a valid destination site" % (destination), **kwargs)
            destSites = ";".join(destination)
            self._addParameter(self.workflow, "Site", "JDL", destSites, description)
        else:
            return self._reportError("Invalid destination site, expected string or list", **kwargs)
        return S_OK()

    #############################################################################
    def setNumberOfProcessors(self, numberOfProcessors=None, minNumberOfProcessors=None, maxNumberOfProcessors=None):
        """Helper function.

        Example usage:

        >>> job = Job()
        >>> job.setNumberOfProcessors(numberOfProcessors=2)
        means that the job needs 2 processors

        >>> job = Job()
        >>> job.setNumberOfProcessors(minNumberOfProcessors=4, maxNumberOfProcessors=8)
        means that the job needs at least 4 processors, and that will use at most 8 processors

        >>> job = Job()
        >>> job.setNumberOfProcessors(minNumberOfProcessors=2)
        means that the job needs at least 2 processors, and that will use all the processors available

        >>> job = Job()
        >>> job.setNumberOfProcessors(minNumberOfProcessors=1)
        means that the job can run in SP mode, and that will use all the processors available
        (so the job could run MP, but also SP)

        >>> job = Job()
        >>> job.setNumberOfProcessors(maxNumberOfProcessors=4)
        is equivalent to
        >>> job.setNumberOfProcessors(minNumberOfProcessors=1, maxNumberOfProcessors=4)
        and it means that the job can run in SP mode, and that will use at most 4 processors
        (so the job could run MP, but also SP)

        >>> job = Job()
        >>> job.setNumberOfProcessors(minNumberOfProcessors=6, maxNumberOfProcessors=4)
        is a non-sense, and will lead to consider that the job can run exactly on 4 processors

        >>> job = Job()
        >>> job.setNumberOfProcessors(numberOfProcessors=3, maxNumberOfProcessors=4)
        will lead to ignore the second parameter

        >>> job = Job()
        >>> job.setNumberOfProcessors(numberOfProcessors=3, minNumberOfProcessors=2)
        will lead to ignore the second parameter

        :param int processors: number of processors required by the job (exact number, unless a min/max are set)
        :param int minNumberOfProcessors: optional min number of processors the job applications can use
        :param int maxNumberOfProcessors: optional max number of processors the job applications can use

        :return: S_OK/S_ERROR
        """
        if numberOfProcessors:
            if not minNumberOfProcessors:
                nProc = numberOfProcessors
            else:
                nProc = max(numberOfProcessors, minNumberOfProcessors)
            if nProc > 1:
                self._addParameter(
                    self.workflow, "NumberOfProcessors", "JDL", nProc, "Exact number of processors requested"
                )
                self._addParameter(
                    self.workflow,
                    "MaxNumberOfProcessors",
                    "JDL",
                    nProc,
                    "Max Number of processors the job applications may use",
                )
            return S_OK()

        if maxNumberOfProcessors and not minNumberOfProcessors:
            minNumberOfProcessors = 1

        if minNumberOfProcessors and maxNumberOfProcessors and minNumberOfProcessors >= maxNumberOfProcessors:
            minNumberOfProcessors = maxNumberOfProcessors

        if (
            minNumberOfProcessors
            and maxNumberOfProcessors
            and minNumberOfProcessors == maxNumberOfProcessors
            and minNumberOfProcessors > 1
        ):
            self._addParameter(
                self.workflow,
                "NumberOfProcessors",
                "JDL",
                minNumberOfProcessors,
                "Exact number of processors requested",
            )
            self._addParameter(
                self.workflow,
                "MaxNumberOfProcessors",
                "JDL",
                minNumberOfProcessors,
                "Max Number of processors the job applications may use",
            )
            return S_OK()

        # By this point there should be a min
        self._addParameter(
            self.workflow,
            "MinNumberOfProcessors",
            "JDL",
            minNumberOfProcessors,
            "Min Number of processors the job applications may use",
        )

        # If not set, will be "all"
        if maxNumberOfProcessors:
            self._addParameter(
                self.workflow,
                "MaxNumberOfProcessors",
                "JDL",
                maxNumberOfProcessors,
                "Max Number of processors the job applications may use",
            )

        return S_OK()

    #############################################################################
    def setDestinationCE(self, ceName, diracSite=None):
        """Developer function.

        Allows to direct a job to a particular Grid CE.
        """
        kwargs = {"ceName": ceName}

        if not diracSite:
            res = getCESiteMapping(ceName)
            if not res["OK"]:
                return self._reportError(res["Message"], **kwargs)
            if not res["Value"]:
                return self._reportError("No DIRAC site name found for CE %s" % (ceName), **kwargs)
            diracSite = res["Value"][ceName]

        self.setDestination(diracSite)
        self._addJDLParameter("GridCE", ceName)
        return S_OK()

    #############################################################################
    def setBannedSites(self, sites):
        """Helper function.

        Can specify a desired destination site for job.  This can be useful
        for debugging purposes but often limits the possible candidate sites
        and overall system response time.

        Example usage:

        >>> job = Job()
        >>> job.setBannedSites(['LCG.GRIDKA.de','LCG.CNAF.it'])

        :param sites: single site string or list
        :type sites: str or python:list
        """
        if isinstance(sites, list) and sites:
            bannedSites = ";".join(sites)
            description = "List of sites excluded by user"
            self._addParameter(self.workflow, "BannedSites", "JDL", bannedSites, description)
        elif isinstance(sites, six.string_types):
            description = "Site excluded by user"
            self._addParameter(self.workflow, "BannedSites", "JDL", sites, description)
        else:
            kwargs = {"sites": sites}
            return self._reportError("Expected site string or list of sites", **kwargs)
        return S_OK()

    #############################################################################
    def setOwner(self, ownerProvided):
        """Developer function.

        Normally users should always specify their immutable DIRAC nickname.
        """
        if not isinstance(ownerProvided, six.string_types):
            return self._reportError("Expected string for owner", **{"ownerProvided": ownerProvided})

        self._addParameter(self.workflow, "Owner", "JDL", ownerProvided, "User specified ID")
        return S_OK()

    #############################################################################
    def setOwnerGroup(self, ownerGroup):
        """Developer function.

        Allows to force expected owner group of proxy.
        """
        if not isinstance(ownerGroup, six.string_types):
            return self._reportError("Expected string for job owner group", **{"ownerGroup": ownerGroup})

        self._addParameter(self.workflow, "OwnerGroup", "JDL", ownerGroup, "User specified owner group.")
        return S_OK()

    #############################################################################
    def setOwnerDN(self, ownerDN):
        """Developer function.

        Allows to force expected owner DN of proxy.
        """
        if not isinstance(ownerDN, six.string_types):
            return self._reportError("Expected string for job owner DN", **{"ownerGroup": ownerDN})

        self._addParameter(self.workflow, "OwnerDN", "JDL", ownerDN, "User specified owner DN.")
        return S_OK()

    #############################################################################
    def setType(self, jobType):
        """Developer function.

        Specify job type for testing purposes.
        """
        if not isinstance(jobType, six.string_types):
            return self._reportError("Expected string for job type", **{"jobType": jobType})

        self._addParameter(self.workflow, "JobType", "JDL", jobType, "User specified type")
        self.type = jobType
        return S_OK()

    #############################################################################
    def setTag(self, tags):
        """Set the Tags job requirements

        Example usage:

        >>> job = Job()
        >>> job.setTag( ['WholeNode','8GBMemory'] )

        :param tags: single tag string or a list of tags
        :type tags: str or python:list
        """

        if isinstance(tags, six.string_types):
            tagValue = tags
        elif isinstance(tags, list):
            tagValue = ";".join(tags)
        else:
            return self._reportError("Expected string or list for job tags", tags=tags)

        self._addParameter(self.workflow, "Tags", "JDL", tagValue, "User specified job tags")
        return S_OK()

    #############################################################################
    def setJobGroup(self, jobGroup):
        """Helper function.

        Allows to group certain jobs according to an ID.

        Example usage:

        >>> job = Job()
        >>> job.setJobGroup('Bs2JPsiPhi')

        :param jobGroup: JobGroup name
        :type jobGroup: string
        """
        if not isinstance(jobGroup, six.string_types):
            return self._reportError("Expected string for job group name", **{"jobGroup": jobGroup})

        description = "User specified job group"
        self._addParameter(self.workflow, "JobGroup", "JDL", jobGroup, description)
        return S_OK()

    #############################################################################
    def setLogLevel(self, logLevel):
        """Helper function.

        Optionally specify a DIRAC logging level for the job, e.g.
        ALWAYS, INFO, VERBOSE, WARN, DEBUG
        by default this is set to the info level.

        Example usage:

        >>> job = Job()
        >>> job.setLogLevel('debug')

        :param logLevel: Logging level
        :type logLevel: string
        """
        kwargs = {"logLevel": logLevel}
        if isinstance(logLevel, six.string_types):
            if logLevel.upper() in gLogger.getAllPossibleLevels():
                description = "User specified logging level"
                self.logLevel = logLevel
                self._addParameter(self.workflow, "LogLevel", "JDL", logLevel, description)
            else:
                return self._reportError('Error Level "%s" not valid' % logLevel, **kwargs)
        else:
            return self._reportError("Expected string for logging level", **kwargs)
        return S_OK()

    #############################################################################
    def setConfigArgs(self, cfgString):
        """Developer function. Allow to pass arbitrary settings to the payload
        configuration service environment.
        """
        if not isinstance(cfgString, six.string_types):
            return self._reportError("Expected string for DIRAC Job Config Args", **{"cfgString": cfgString})

        description = "User specified cfg settings"
        self._addParameter(self.workflow, "JobConfigArgs", "JDL", cfgString, description)
        return S_OK()

    #############################################################################
    def setExecutionEnv(self, environmentDict):
        """Helper function.

        Optionally specify a dictionary of key, value pairs to be set before
        the job executes e.g. {'MYVAR':3}

        The standard application environment variables are always set so this
        is intended for user variables only.

        Example usage:

        >>> job = Job()
        >>> job.setExecutionEnv({'<MYVARIABLE>':'<VALUE>'})

        :param environmentDict: Environment variables
        :type environmentDict: dictionary
        """
        kwargs = {"environmentDict": environmentDict}
        if not isinstance(environmentDict, dict):
            return self._reportError("Expected dictionary of environment variables", **kwargs)

        if environmentDict:
            environment = []
            for var, val in environmentDict.items():
                try:
                    environment.append("=".join([str(var), urlquote(str(val))]))
                except Exception:
                    return self._reportError("Expected string for environment variable key value pairs", **kwargs)

            envStr = ";".join(environment)
            description = "Env vars specified by user"
            self._addParameter(self.workflow, "ExecutionEnvironment", "JDL", envStr, description)
        return S_OK()

    #############################################################################
    def execute(self):
        """Developer function. Executes the job locally."""
        self.workflow.createCode()
        self.workflow.execute()

    #############################################################################
    def _getParameters(self):
        """Developer function.
        Method to return the workflow parameters.
        """
        wfParams = {}
        params = self.workflow.parameters
        for par in params:
            wfParams[par.getName()] = par.getValue()
        return wfParams

    #############################################################################

    def __setJobDefaults(self):
        """Set job default values. Note that the system configuration is set to "ANY"."""
        self._addParameter(self.workflow, "JobType", "JDL", self.type, "Job Type")
        self._addParameter(self.workflow, "Priority", "JDL", self.priority, "User Job Priority")
        self._addParameter(self.workflow, "JobGroup", "JDL", self.group, "Name of the JobGroup")
        self._addParameter(self.workflow, "JobName", "JDL", self.name, "Name of Job")
        self._addParameter(self.workflow, "StdOutput", "JDL", self.stdout, "Standard output file")
        self._addParameter(self.workflow, "StdError", "JDL", self.stderr, "Standard error file")
        self._addParameter(self.workflow, "InputData", "JDL", "", "Default null input data value")
        self._addParameter(self.workflow, "LogLevel", "JDL", self.logLevel, "Job Logging Level")
        self._addParameter(self.workflow, "arguments", "string", "", "Arguments to executable Step")
        # Those 2 below are need for on-site resolution
        self._addParameter(
            self.workflow, "ParametricInputData", "string", "", "Default null parametric input data value"
        )
        self._addParameter(
            self.workflow, "ParametricInputSandbox", "string", "", "Default null parametric input sandbox value"
        )

    #############################################################################

    @staticmethod
    def _addParameter(wObject, name, ptype, value, description, io="input"):
        """Internal Function

        Adds a parameter to the object.
        """
        if io == "input":
            inBool = True
            outBool = False
        elif io == "output":
            inBool = False
            outBool = True
        else:
            raise TypeError("I/O flag is either input or output")

        par = Parameter(name, value, ptype, "", "", inBool, outBool, description)
        wObject.addParameter(Parameter(parameter=par))

    ############################################################################
    def _resolveInputSandbox(self, inputSandbox):
        """Internal function.

        Resolves wildcards for input sandbox files.  This is currently linux
        specific and should be modified.
        """
        resolvedIS = []
        for i in inputSandbox:
            if not re.search(r"\*", i):
                if not os.path.isdir(i):
                    resolvedIS.append(i)

        for name in inputSandbox:
            if re.search(r"\*", name):  # escape the star character...
                cmd = "ls -d " + name
                output = systemCall(10, shlex.split(cmd))
                if not output["OK"]:
                    self.log.error("Could not perform: ", cmd)
                elif output["Value"][0]:
                    self.log.error(" Failed getting the files ", output["Value"][2])
                else:
                    files = output["Value"][1].split()
                    for check in files:
                        if os.path.isfile(check):
                            self.log.verbose("Found file " + check + " appending to Input Sandbox")
                            resolvedIS.append(check)
                        if os.path.isdir(check):
                            if re.search("/$", check):  # users can specify e.g. /my/dir/lib/
                                check = check[:-1]
                            tarName = os.path.basename(check)
                            directory = os.path.dirname(check)  # if just the directory this is null
                            if directory:
                                cmd = "tar cfz " + tarName + ".tar.gz " + " -C " + directory + " " + tarName
                            else:
                                cmd = "tar cfz " + tarName + ".tar.gz " + tarName

                            output = systemCall(60, shlex.split(cmd))
                            if not output["OK"]:
                                self.log.error("Could not perform: %s" % (cmd))
                            resolvedIS.append(tarName + ".tar.gz")
                            self.log.verbose(
                                "Found directory " + check + ", appending " + check + ".tar.gz to Input Sandbox"
                            )

            if os.path.isdir(name):
                self.log.verbose(
                    "Found specified directory " + name + ", appending " + name + ".tar.gz to Input Sandbox"
                )
                if re.search("/$", name):  # users can specify e.g. /my/dir/lib/
                    name = name[:-1]
                tarName = os.path.basename(name)
                directory = os.path.dirname(name)  # if just the directory this is null
                if directory:
                    cmd = "tar cfz " + tarName + ".tar.gz " + " -C " + directory + " " + tarName
                else:
                    cmd = "tar cfz " + tarName + ".tar.gz " + tarName

                output = systemCall(60, shlex.split(cmd))
                if not output["OK"]:
                    self.log.error("Could not perform: %s" % (cmd))
                else:
                    resolvedIS.append(tarName + ".tar.gz")

        return resolvedIS

    #############################################################################

    def _toXML(self):
        """Returns an XML representation of itself as a Job."""
        return self.workflow.toXML()

    def _handleParameterSequences(self, paramsDict, arguments):

        for pName in self.parameterSeqs:
            if pName in paramsDict:
                if pName == "InputSandbox":
                    if isinstance(paramsDict[pName]["value"], list):
                        paramsDict[pName]["value"].append("%%(%s)s" % pName)
                    elif isinstance(paramsDict[pName]["value"], six.string_types):
                        if paramsDict[pName]["value"]:
                            paramsDict[pName]["value"] += ";%%(%s)s" % pName
                        else:
                            paramsDict[pName]["value"] = "%%(%s)s" % pName
                elif "jdl" in paramsDict[pName]["type"].lower():
                    # If a parameter with the same name as the sequence name already exists
                    # and is a list, then extend it by the sequence value. If it is not a
                    # list, then replace it by the sequence value
                    if isinstance(paramsDict[pName]["value"], list):
                        currentParams = paramsDict[pName]["value"]
                        tmpList = []
                        pData = self.parameterSeqs[pName]
                        if isinstance(pData[0], list):
                            for pElement in pData:
                                tmpList.append(currentParams + pElement)
                        else:
                            for pElement in pData:
                                tmpList.append(currentParams + [pElement])
                        self.parameterSeqs[pName] = tmpList
                    paramsDict[pName]["value"] = "%%(%s)s" % pName
            else:
                paramsDict[pName] = {}
                paramsDict[pName]["type"] = "JDL"
                paramsDict[pName]["value"] = "%%(%s)s" % pName

            paramsDict["Parameters.%s" % pName] = {}
            paramsDict["Parameters.%s" % pName]["value"] = self.parameterSeqs[pName]
            paramsDict["Parameters.%s" % pName]["type"] = "JDL"
            if pName in self.wfArguments:
                arguments.append(" -p %s=%%(%s)s" % (self.wfArguments[pName], pName))

        return paramsDict, arguments

    #############################################################################
    def _toJDL(self, xmlFile="", jobDescriptionObject=None):
        """Creates a JDL representation of itself as a Job.

        Example usage:

        >>> job = Job()
        >>> job._toJDL()

        :param xmlFile: location of the XML file
        :type xmlFile: str
        :param jobDescriptionObject: if not None, it must be a StringIO object
        :type jobDescriptionObject: StringIO

        :returns: JDL (str)
        """
        # Check if we have to do old bootstrap...
        classadJob = ClassAd("[]")

        paramsDict = {}
        params = self.workflow.parameters  # ParameterCollection object

        paramList = params
        for param in paramList:
            paramsDict[param.getName()] = {"type": param.getType(), "value": param.getValue()}

        arguments = []
        scriptName = "jobDescription.xml"

        if jobDescriptionObject is None:
            # if we are here it's because there's a real file, on disk, that is named 'jobDescription.xml'
            # Messy but need to account for xml file being in /tmp/guid dir
            if self.script:
                if os.path.exists(self.script):
                    scriptName = os.path.abspath(self.script)
                    self.log.verbose("Found script name %s" % scriptName)
                else:
                    self.log.warn("File not found", self.script)
            else:
                if xmlFile:
                    if os.path.exists(xmlFile):
                        self.log.verbose("Found XML File %s" % xmlFile)
                        scriptName = xmlFile
                    else:
                        self.log.warn("File not found", xmlFile)
                else:
                    if os.path.exists("jobDescription.xml"):
                        scriptName = os.path.abspath("jobDescription.xml")
                        self.log.verbose("Found script name %s" % scriptName)
                    else:
                        self.log.warn("Job description XML file not found")
            self.addToInputSandbox.append(scriptName)

        elif isinstance(jobDescriptionObject, StringIO):
            self.log.verbose("jobDescription is passed in as a StringIO object")

        else:
            self.log.error("Where's the job description?")

        arguments.append(os.path.basename(scriptName))
        if "LogLevel" in paramsDict:
            if paramsDict["LogLevel"]["value"]:
                arguments.append("-o LogLevel=%s" % (paramsDict["LogLevel"]["value"]))
            else:
                self.log.warn("Job LogLevel defined with null value")
        if "DIRACSetup" in paramsDict:
            if paramsDict["DIRACSetup"]["value"]:
                arguments.append("-o DIRAC/Setup=%s" % (paramsDict["DIRACSetup"]["value"]))
            else:
                self.log.warn("Job DIRACSetup defined with null value")
        if "JobConfigArgs" in paramsDict:
            if paramsDict["JobConfigArgs"]["value"]:
                arguments.append("--cfg %s" % (paramsDict["JobConfigArgs"]["value"]))
            else:
                self.log.warn("JobConfigArgs defined with null value")
        if self.parametricWFArguments:
            for name, value in self.parametricWFArguments.items():
                arguments.append("-p %s='%s'" % (name, value))

        classadJob.insertAttributeString("Executable", self.executable)
        self.addToOutputSandbox.append(self.stderr)
        self.addToOutputSandbox.append(self.stdout)

        # Extract i/o sandbox parameters from steps and any input data parameters
        # to do when introducing step-level api...

        # To add any additional files to input and output sandboxes
        if self.addToInputSandbox:
            extraFiles = ";".join(self.addToInputSandbox)
            if "InputSandbox" in paramsDict:
                currentFiles = paramsDict["InputSandbox"]["value"]
                finalInputSandbox = currentFiles + ";" + extraFiles
                uniqueInputSandbox = uniqueElements(finalInputSandbox.split(";"))
                paramsDict["InputSandbox"]["value"] = ";".join(uniqueInputSandbox)
                self.log.verbose("Final unique Input Sandbox %s" % (";".join(uniqueInputSandbox)))
            else:
                paramsDict["InputSandbox"] = {}
                paramsDict["InputSandbox"]["value"] = extraFiles
                paramsDict["InputSandbox"]["type"] = "JDL"

        if self.addToOutputSandbox:
            extraFiles = ";".join(self.addToOutputSandbox)
            if "OutputSandbox" in paramsDict:
                currentFiles = paramsDict["OutputSandbox"]["value"]
                finalOutputSandbox = currentFiles + ";" + extraFiles
                uniqueOutputSandbox = uniqueElements(finalOutputSandbox.split(";"))
                paramsDict["OutputSandbox"]["value"] = ";".join(uniqueOutputSandbox)
                self.log.verbose("Final unique Output Sandbox %s" % (";".join(uniqueOutputSandbox)))
            else:
                paramsDict["OutputSandbox"] = {}
                paramsDict["OutputSandbox"]["value"] = extraFiles
                paramsDict["OutputSandbox"]["type"] = "JDL"

        if self.addToInputData:
            extraFiles = ";".join(self.addToInputData)
            if "InputData" in paramsDict:
                currentFiles = paramsDict["InputData"]["value"]
                finalInputData = extraFiles
                if currentFiles:
                    finalInputData = currentFiles + ";" + extraFiles
                uniqueInputData = uniqueElements(finalInputData.split(";"))
                paramsDict["InputData"]["value"] = ";".join(uniqueInputData)
                self.log.verbose("Final unique Input Data %s" % (";".join(uniqueInputData)))
            else:
                paramsDict["InputData"] = {}
                paramsDict["InputData"]["value"] = extraFiles
                paramsDict["InputData"]["type"] = "JDL"

        # Handle parameter sequences
        if self.numberOfParameters > 0:
            paramsDict, arguments = self._handleParameterSequences(paramsDict, arguments)

        classadJob.insertAttributeString("Arguments", " ".join(arguments))

        # Add any JDL parameters to classad obeying lists with ';' rule
        for name, props in paramsDict.items():
            ptype = props["type"]
            value = props["value"]
            if isinstance(value, six.string_types) and re.search(";", value):
                value = value.split(";")
            if name.lower() == "requirements" and ptype == "JDL":
                self.log.verbose("Found existing requirements: %s" % (value))

            if re.search("^JDL", ptype):
                if isinstance(value, list):
                    if isinstance(value[0], list):
                        classadJob.insertAttributeVectorStringList(name, value)
                    else:
                        classadJob.insertAttributeVectorInt(name, value)
                elif isinstance(value, six.string_types) and value:
                    classadJob.insertAttributeInt(name, value)
                elif isinstance(value, six.integer_types + (float,)):
                    classadJob.insertAttributeInt(name, value)

        if self.numberOfParameters > 0:
            classadJob.insertAttributeInt("Parameters", self.numberOfParameters)

        for fToBeRemoved in [scriptName, self.stdout, self.stderr]:
            try:
                self.addToInputSandbox.remove(fToBeRemoved)
            except ValueError:
                pass

        jdl = classadJob.asJDL()
        start = jdl.find("[")
        end = jdl.rfind("]")
        return jdl[(start + 1) : (end - 1)]

    #############################################################################
    def _setParamValue(self, name, value):
        """Internal Function. Sets a parameter value, used for production."""
        return self.workflow.setValue(name, value)

    #############################################################################
    def _addJDLParameter(self, name, value):
        """Developer function, add an arbitrary JDL parameter."""
        self._addParameter(self.workflow, name, "JDL", value, "Optional JDL parameter added")
        return self.workflow.setValue(name, value)

    #############################################################################

    def runLocal(self, dirac=None):
        """The dirac (API) object is for local submission."""

        if dirac is None:
            dirac = Dirac()

        return dirac.submitJob(self, mode="local")


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



""" The mind is a service the distributes "task" to executors
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pprint

import six
from DIRAC import gLogger
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR, isReturnStructure
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities.ExecutorDispatcher import ExecutorDispatcher, ExecutorDispatcherCallbacks


class ExecutorMindHandler(RequestHandler):

    MSG_DEFINITIONS = {
        "ProcessTask": {"taskId": six.integer_types, "taskStub": six.string_types, "eType": six.string_types},
        "TaskDone": {"taskId": six.integer_types, "taskStub": six.string_types},
        "TaskFreeze": {
            "taskId": six.integer_types + (str,),
            "taskStub": six.string_types,
            "freezeTime": six.integer_types,
        },
        "TaskError": {
            "taskId": six.integer_types,
            "errorMsg": six.string_types,
            "taskStub": six.string_types,
            "eType": six.string_types,
        },
        "ExecutorError": {"taskId": six.integer_types, "errorMsg": six.string_types, "eType": six.string_types},
    }

    class MindCallbacks(ExecutorDispatcherCallbacks):
        def __init__(self, sendTaskCB, dispatchCB, disconnectCB, taskProcCB, taskFreezeCB, taskErrCB):
            self.__sendTaskCB = sendTaskCB
            self.__dispatchCB = dispatchCB
            self.__disconnectCB = disconnectCB
            self.__taskProcDB = taskProcCB
            self.__taskFreezeCB = taskFreezeCB
            self.__taskErrCB = taskErrCB
            self.__allowedClients = []

        def cbSendTask(self, taskId, taskObj, eId, eType):
            return self.__sendTaskCB(taskId, taskObj, eId, eType)

        def cbDispatch(self, taskId, taskObj, pathExecuted):
            return self.__dispatchCB(taskId, taskObj, pathExecuted)

        def cbDisconectExecutor(self, eId):
            return self.__disconnectCB(eId)

        def cbTaskError(self, taskId, taskObj, errorMsg):
            return self.__taskErrCB(taskId, taskObj, errorMsg)

        def cbTaskProcessed(self, taskId, taskObj, eType):
            return self.__taskProcDB(taskId, taskObj, eType)

        def cbTaskFreeze(self, taskId, taskObj, eType):
            return self.__taskFreezeCB(taskId, taskObj, eType)

    ###
    # End of callbacks
    ###

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        gLogger.notice("Initializing Executor dispatcher")
        cls.__eDispatch = ExecutorDispatcher(cls.srv_getMonitor())
        cls.__callbacks = ExecutorMindHandler.MindCallbacks(
            cls.__sendTask,
            cls.exec_dispatch,
            cls.__execDisconnected,
            cls.exec_taskProcessed,
            cls.exec_taskFreeze,
            cls.exec_taskError,
        )
        cls.__eDispatch.setCallbacks(cls.__callbacks)
        cls.__allowedClients = []
        if cls.log.shown("VERBOSE"):
            gThreadScheduler.setMinValidPeriod(1)
            gThreadScheduler.addPeriodicTask(
                10,
                lambda: cls.log.verbose(
                    "== Internal state ==",
                    "\n%s\n===========" % pprint.pformat(cls.__eDispatch._internals()),
                ),
            )
        return S_OK()

    @classmethod
    def setAllowedClients(cls, aClients):
        if not isinstance(aClients, (list, tuple)):
            aClients = (aClients,)
        cls.__allowedClients = aClients

    @classmethod
    def __sendTask(self, taskId, taskObj, eId, eType):
        try:
            result = self.exec_prepareToSend(taskId, taskObj, eId)
            if not result["OK"]:
                return result
        except Exception as excp:
            gLogger.exception("Exception while executing prepareToSend:", taskId, lException=excp)
            return S_ERROR("Cannot presend task")
        try:
            result = self.exec_serializeTask(taskObj)
        except Exception as excp:
            gLogger.exception("Exception while serializing task", taskId, lException=excp)
            return S_ERROR("Cannot serialize task %s: %s" % (taskId, str(excp)))
        if not isReturnStructure(result):
            raise Exception("exec_serializeTask does not return a return structure")
        if not result["OK"]:
            return result
        taskStub = result["Value"]
        result = self.srv_msgCreate("ProcessTask")
        if not result["OK"]:
            gLogger.error("Failed to create message for", "%s %r" % (taskId, result))
            return result
        msgObj = result["Value"]
        msgObj.taskId = taskId
        msgObj.taskStub = taskStub
        msgObj.eType = eType
        return self.srv_msgSend(eId, msgObj)

    @classmethod
    def __execDisconnected(cls, trid):
        result = cls.srv_disconnectClient(trid)
        if not result["OK"]:
            return result
        return cls.exec_executorDisconnected(trid)

    auth_conn_new = ["all"]

    def conn_new(self, trid, identity, kwargs):
        if "executorTypes" in kwargs and kwargs["executorTypes"]:
            return S_OK()
        for aClient in self.__allowedClients:
            if aClient in kwargs and kwargs[aClient]:
                return S_OK()
        return S_ERROR("Only executors are allowed to connect")

    auth_conn_connected = ["all"]

    def conn_connected(self, trid, identity, kwargs):
        for aClient in self.__allowedClients:
            if aClient in kwargs and kwargs[aClient]:
                return S_OK()
        try:
            numTasks = max(1, int(kwargs["maxTasks"]))
        except Exception:
            numTasks = 1
        self.__eDispatch.addExecutor(trid, kwargs["executorTypes"])
        return self.exec_executorConnected(trid, kwargs["executorTypes"])

    auth_conn_drop = ["all"]

    def conn_drop(self, trid):
        gLogger.warn("Triggered conn_drop for", trid)
        self.__eDispatch.removeExecutor(trid)
        return self.srv_disconnect(trid)

    auth_msg_TaskDone = ["all"]

    def msg_TaskDone(self, msgObj):
        taskId = msgObj.taskId
        try:
            result = self.exec_deserializeTask(msgObj.taskStub)
        except Exception as excp:
            gLogger.exception("Exception while deserializing task", taskId, lException=excp)
            return S_ERROR("Cannot deserialize task %s: %s" % (taskId, str(excp)))
        if not isReturnStructure(result):
            raise Exception("exec_deserializeTask does not return a return structure")
        if not result["OK"]:
            return result
        taskObj = result["Value"]
        result = self.__eDispatch.taskProcessed(self.srv_getTransportID(), msgObj.taskId, taskObj)
        if not result["OK"]:
            gLogger.error("There was a problem processing task", "%s: %s" % (taskId, result["Message"]))
        return S_OK()

    auth_msg_TaskFreeze = ["all"]

    def msg_TaskFreeze(self, msgObj):
        taskId = msgObj.taskId
        try:
            result = self.exec_deserializeTask(msgObj.taskStub)
        except Exception as excp:
            gLogger.exception("Exception while deserializing task", taskId, lException=excp)
            return S_ERROR("Cannot deserialize task %s: %s" % (taskId, str(excp)))
        if not isReturnStructure(result):
            raise Exception("exec_deserializeTask does not return a return structure")
        if not result["OK"]:
            return result
        taskObj = result["Value"]
        result = self.__eDispatch.freezeTask(self.srv_getTransportID(), msgObj.taskId, msgObj.freezeTime, taskObj)
        if not result["OK"]:
            gLogger.error("There was a problem freezing task", "%s: %s" % (taskId, result["Message"]))
        return S_OK()

    auth_msg_TaskError = ["all"]

    def msg_TaskError(self, msgObj):
        taskId = msgObj.taskId
        try:
            result = self.exec_deserializeTask(msgObj.taskStub)
        except Exception as excp:
            gLogger.exception("Exception while deserializing task", taskId, lException=excp)
            return S_ERROR("Cannot deserialize task %s: %s" % (taskId, str(excp)))
        if not isReturnStructure(result):
            raise Exception("exec_deserializeTask does not return a return structure")
        if not result["OK"]:
            return result
        taskObj = result["Value"]
        # TODO: Check the executor has privileges over the task
        self.__eDispatch.removeTask(msgObj.taskId)
        try:
            self.exec_taskError(msgObj.taskId, taskObj, msgObj.errorMsg)
        except Exception as excp:
            gLogger.exception("Exception when processing task", msgObj.taskId, lException=excp)
        return S_OK()

    auth_msg_ExecutorError = ["all"]

    def msg_ExecutorError(self, msgObj):
        gLogger.info("Disconnecting executor by error:", msgObj.errorMsg)
        self.__eDispatch.removeExecutor(self.srv_getTransportID())
        return self.srv_disconnect()

    #######
    # Utilities functions
    #######

    @classmethod
    def getTaskIds(cls):
        return cls.__eDispatch.getTaskIds()

    @classmethod
    def getExecutorsConnected(cls):
        return cls.__eDispatch.getExecutorsConnected()

    @classmethod
    def setFailedOnTooFrozen(cls, value):
        # If a task is frozen too many times, send error or forget task?
        cls.__eDispatch.setFailedOnTooFrozen(value)

    @classmethod
    def setFreezeOnFailedDispatch(cls, value):
        # If a task fails to properly dispatch, freeze it?
        cls.__eDispatch.setFreezeOnFailedDispatch(value)

    @classmethod
    def setFreezeOnUnknownExecutor(cls, value):
        # If a task needs to go to an executor that has not connected. Forget the task?
        cls.__eDispatch.setFreezeOnUnknownExecutor(value)

    #######
    # Methods that can be overwritten
    #######

    @classmethod
    def exec_executorDisconnected(cls, trid):
        return S_OK()

    @classmethod
    def exec_executorConnected(cls, execName, trid):
        return S_OK()

    @classmethod
    def exec_prepareToSend(cls, taskId, taskObj, eId):
        return S_OK()

    ########
    #  Methods to be used by the real services
    ########

    @classmethod
    def executeTask(cls, taskId, taskObj):
        return cls.__eDispatch.addTask(taskId, taskObj)

    @classmethod
    def forgetTask(cls, taskId):
        return cls.__eDispatch.removeTask(taskId)

    ########
    #  Methods that need to be overwritten
    ########

    @classmethod
    def exec_dispatch(cls, taskId, taskObj, pathExecuted):
        raise Exception("No exec_dispatch defined or it is not a classmethod!!")

    @classmethod
    def exec_serializeTask(cls, taskObj):
        raise Exception("No exec_serializeTask defined or it is not a classmethod!!")

    @classmethod
    def exec_deserializeTask(cls, taskStub):
        raise Exception("No exec_deserializeTask defined or it is not a classmethod!!")

    @classmethod
    def exec_taskError(cls, taskId, taskObj, errorMsg):
        raise Exception("No exec_taskError defined or it is not a classmethod!!")

    @classmethod
    def exec_taskProcessed(cls, taskId, taskObj, eType):
        raise Exception("No exec_taskProcessed defined or it is not a classmethod!!")

    @classmethod
    def exec_taskFreeze(cls, taskId, taskObj, eType):
        return S_OK()



""" Handler for CAs + CRLs bundles
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import tarfile
import os
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC import gLogger, S_OK, S_ERROR, gConfig
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler
from DIRAC.Core.Utilities import File, List
from DIRAC.Core.Security import Locations, Utilities


class BundleManager(object):
    def __init__(self, baseCSPath):
        self.__csPath = baseCSPath
        self.__bundles = {}
        self.updateBundles()

    def __getDirsToBundle(self):
        dirsToBundle = {}
        result = gConfig.getOptionsDict("%s/DirsToBundle" % self.__csPath)
        if result["OK"]:
            dB = result["Value"]
            for bId in dB:
                dirsToBundle[bId] = List.fromChar(dB[bId])
        if gConfig.getValue("%s/BundleCAs" % self.__csPath, True):
            dirsToBundle["CAs"] = [
                "%s/*.0" % Locations.getCAsLocation(),
                "%s/*.signing_policy" % Locations.getCAsLocation(),
                "%s/*.pem" % Locations.getCAsLocation(),
            ]
        if gConfig.getValue("%s/BundleCRLs" % self.__csPath, True):
            dirsToBundle["CRLs"] = ["%s/*.r0" % Locations.getCAsLocation()]
        return dirsToBundle

    def getBundles(self):
        return dict([(bId, self.__bundles[bId]) for bId in self.__bundles])

    def bundleExists(self, bId):
        return bId in self.__bundles

    def getBundleVersion(self, bId):
        try:
            return self.__bundles[bId][0]
        except Exception:
            return ""

    def getBundleData(self, bId):
        try:
            return self.__bundles[bId][1]
        except Exception:
            return ""

    def updateBundles(self):
        dirsToBundle = self.__getDirsToBundle()
        # Delete bundles that don't have to be updated
        for bId in self.__bundles:
            if bId not in dirsToBundle:
                gLogger.info("Deleting old bundle %s" % bId)
                del self.__bundles[bId]
        for bId in dirsToBundle:
            bundlePaths = dirsToBundle[bId]
            gLogger.info("Updating %s bundle %s" % (bId, bundlePaths))
            buffer_ = six.BytesIO()
            filesToBundle = sorted(File.getGlobbedFiles(bundlePaths))
            if filesToBundle:
                commonPath = os.path.commonprefix(filesToBundle)
                commonEnd = len(commonPath)
                gLogger.info("Bundle will have %s files with common path %s" % (len(filesToBundle), commonPath))
                with tarfile.open("dummy", "w:gz", buffer_) as tarBuffer:
                    for filePath in filesToBundle:
                        tarBuffer.add(filePath, filePath[commonEnd:])
                zippedData = buffer_.getvalue()
                buffer_.close()
                hash_ = File.getMD5ForFiles(filesToBundle)
                gLogger.info("Bundled %s : %s bytes (%s)" % (bId, len(zippedData), hash_))
                self.__bundles[bId] = (hash_, zippedData)
            else:
                self.__bundles[bId] = (None, None)


class BundleDeliveryHandler(RequestHandler):
    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        csPath = serviceInfoDict["serviceSectionPath"]
        cls.bundleManager = BundleManager(csPath)
        updateBundleTime = gConfig.getValue("%s/BundlesLifeTime" % csPath, 3600 * 6)
        gLogger.info("Bundles will be updated each %s secs" % updateBundleTime)
        gThreadScheduler.addPeriodicTask(updateBundleTime, cls.bundleManager.updateBundles)
        return S_OK()

    types_getListOfBundles = []

    @classmethod
    def export_getListOfBundles(cls):
        return S_OK(cls.bundleManager.getBundles())

    def transfer_toClient(self, fileId, token, fileHelper):
        version = ""
        if isinstance(fileId, six.string_types):
            if fileId in ["CAs", "CRLs"]:
                return self.__transferFile(fileId, fileHelper)
            else:
                bId = fileId
        elif isinstance(fileId, (list, tuple)):
            if len(fileId) == 0:
                fileHelper.markAsTransferred()
                return S_ERROR("No bundle specified!")
            elif len(fileId) == 1:
                bId = fileId[0]
            else:
                bId = fileId[0]
                version = fileId[1]
        if not self.bundleManager.bundleExists(bId):
            fileHelper.markAsTransferred()
            return S_ERROR("Unknown bundle %s" % bId)

        bundleVersion = self.bundleManager.getBundleVersion(bId)
        if bundleVersion is None:
            fileHelper.markAsTransferred()
            return S_ERROR("Empty bundle %s" % bId)

        if version == bundleVersion:
            fileHelper.markAsTransferred()
            return S_OK(bundleVersion)

        buffer_ = six.BytesIO(self.bundleManager.getBundleData(bId))
        result = fileHelper.DataSourceToNetwork(buffer_)
        buffer_.close()
        if not result["OK"]:
            return result
        return S_OK(bundleVersion)

    def __transferFile(self, filetype, fileHelper):
        """
        This file is creates and transfers the CAs or CRLs file to the client.
        :param str filetype: we can define which file will be transfered to the client
        :param object fileHelper:
        :return: S_OK or S_ERROR
        """
        if filetype == "CAs":
            retVal = Utilities.generateCAFile()
        elif filetype == "CRLs":
            retVal = Utilities.generateRevokedCertsFile()
        else:
            return S_ERROR("Not supported file type %s" % filetype)

        if not retVal["OK"]:
            return retVal
        else:
            result = fileHelper.getFileDescriptor(retVal["Value"], "r")
            if not result["OK"]:
                result = fileHelper.sendEOF()
                # better to check again the existence of the file
                if not os.path.exists(retVal["Value"]):
                    return S_ERROR("File %s does not exist" % os.path.basename(retVal["Value"]))
                else:
                    return S_ERROR("Failed to get file descriptor")
            fileDescriptor = result["Value"]
            result = fileHelper.FDToNetwork(fileDescriptor)
            fileHelper.oFile.close()  # close the file and return
            return result



""" DIRAC Transformation DB

    Transformation database is used to collect and serve the necessary information
    in order to automate the task of job preparation for high level transformations.
    This class is typically used as a base class for more specific data processing
    databases
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import re
import time
import threading

from errno import ENOENT

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Base.DB import DB
from DIRAC.Core.Utilities.DErrno import cmpError
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Core.Utilities.List import stringListToString, intListToString, breakListIntoChunks
from DIRAC.Core.Utilities.Shifter import setupShifterProxyInEnv
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities.Subprocess import pythonCall
from DIRAC.DataManagementSystem.Client.MetaQuery import MetaQuery

__RCSID__ = "$Id$"

MAX_ERROR_COUNT = 10

#############################################################################


class TransformationDB(DB):
    """TransformationDB class"""

    def __init__(self, dbname=None, dbconfig=None, dbIn=None):
        """The standard constructor takes the database name (dbname) and the name of the
        configuration section (dbconfig)
        """

        if not dbname:
            dbname = "TransformationDB"
        if not dbconfig:
            dbconfig = "Transformation/TransformationDB"

        if not dbIn:
            DB.__init__(self, dbname, dbconfig)

        self.lock = threading.Lock()

        self.allowedStatusForTasks = ("Unused", "ProbInFC")

        self.TRANSPARAMS = [
            "TransformationID",
            "TransformationName",
            "Description",
            "LongDescription",
            "CreationDate",
            "LastUpdate",
            "AuthorDN",
            "AuthorGroup",
            "Type",
            "Plugin",
            "AgentType",
            "Status",
            "FileMask",
            "TransformationGroup",
            "GroupSize",
            "InheritedFrom",
            "Body",
            "MaxNumberOfTasks",
            "EventsPerTask",
            "TransformationFamily",
        ]

        self.mutable = [
            "TransformationName",
            "Description",
            "LongDescription",
            "AgentType",
            "Status",
            "GroupSize",
            "MaxNumberOfTasks",
            "TransformationFamily",
            "Body",
        ]  # for the moment include TransformationFamily

        self.TRANSFILEPARAMS = [
            "TransformationID",
            "FileID",
            "Status",
            "TaskID",
            "TargetSE",
            "UsedSE",
            "ErrorCount",
            "LastUpdate",
            "InsertedTime",
        ]

        self.TRANSFILETASKPARAMS = ["TransformationID", "FileID", "TaskID"]

        self.TASKSPARAMS = [
            "TaskID",
            "TransformationID",
            "ExternalStatus",
            "ExternalID",
            "TargetSE",
            "CreationTime",
            "LastUpdateTime",
        ]

        self.ADDITIONALPARAMETERS = ["TransformationID", "ParameterName", "ParameterValue", "ParameterType"]

        # Intialize filter Queries with Input Meta Queries
        self.filterQueries = []
        res = self.__updateFilterQueries()
        if not res["OK"]:
            gLogger.fatal("Failed to create filter queries")

        # This is here to ensure full compatibility between different versions of the MySQL DB schema
        self.isTransformationTasksInnoDB = True
        res = self._query("SELECT Engine FROM INFORMATION_SCHEMA.TABLES WHERE table_name = 'TransformationTasks'")
        if not res["OK"]:
            raise RuntimeError(res["Message"])
        else:
            engine = res["Value"][0][0]
            if engine.lower() != "innodb":
                self.isTransformationTasksInnoDB = False

    def getName(self):
        """Get the database name"""
        return self.dbName

    ###########################################################################
    #
    # These methods manipulate the Transformations table
    #

    def addTransformation(
        self,
        transName,
        description,
        longDescription,
        authorDN,
        authorGroup,
        transType,
        plugin,
        agentType,
        fileMask,
        transformationGroup="General",
        groupSize=1,
        inheritedFrom=0,
        body="",
        maxTasks=0,
        eventsPerTask=0,
        addFiles=True,
        connection=False,
        inputMetaQuery=None,
        outputMetaQuery=None,
    ):
        """Add new transformation definition including its input streams"""
        connection = self.__getConnection(connection)
        res = self._getTransformationID(transName, connection=connection)
        if res["OK"]:
            return S_ERROR(
                "Transformation with name %s already exists with TransformationID = %d" % (transName, res["Value"])
            )
        elif res["Message"] != "Transformation does not exist":
            return res
        self.lock.acquire()
        res = self._escapeString(body)
        if not res["OK"]:
            return S_ERROR("Failed to parse the transformation body")
        body = res["Value"]
        req = "INSERT INTO Transformations (TransformationName,Description,LongDescription, \
                                        CreationDate,LastUpdate,AuthorDN,AuthorGroup,Type,Plugin,AgentType,\
                                        FileMask,Status,TransformationGroup,GroupSize,\
                                        InheritedFrom,Body,MaxNumberOfTasks,EventsPerTask)\
                                VALUES ('%s','%s','%s',\
                                        UTC_TIMESTAMP(),UTC_TIMESTAMP(),'%s','%s','%s','%s','%s',\
                                        '%s','New','%s',%f,\
                                        %d,%s,%d,%d);" % (
            transName,
            description,
            longDescription,
            authorDN,
            authorGroup,
            transType,
            plugin,
            agentType,
            fileMask,
            transformationGroup,
            groupSize,
            inheritedFrom,
            body,
            maxTasks,
            eventsPerTask,
        )
        res = self._update(req, connection)
        if not res["OK"]:
            self.lock.release()
            return res
        transID = res["lastRowId"]
        self.lock.release()

        # Add Input and Output Meta Queries to the transformation if they are defined
        if inputMetaQuery:
            res = self.createTransformationMetaQuery(transID, inputMetaQuery, "Input")
            if not res["OK"]:
                gLogger.error("Failed to add input meta query to the transformation", res["Message"])
                return self.deleteTransformation(transID, connection=connection)
        if outputMetaQuery:
            res = self.createTransformationMetaQuery(transID, outputMetaQuery, "Output")
            if not res["OK"]:
                gLogger.error("Failed to add output meta query to the transformation", res["Message"])
                return self.deleteTransformation(transID, connection=connection)

        # If the transformation has an input data specification
        if inputMetaQuery:
            self.filterQueries.append((transID, inputMetaQuery))

        if inheritedFrom:
            res = self._getTransformationID(inheritedFrom, connection=connection)
            if not res["OK"]:
                gLogger.error("Failed to get ID for parent transformation, now deleting", res["Message"])
                return self.deleteTransformation(transID, connection=connection)
            originalID = res["Value"]
            # FIXME: this is not the right place to change status information, and in general the whole should not be here
            res = self.setTransformationParameter(
                originalID, "Status", "Completing", author=authorDN, connection=connection
            )
            if not res["OK"]:
                gLogger.error("Failed to update parent transformation status: now deleting", res["Message"])
                return self.deleteTransformation(transID, connection=connection)
            res = self.setTransformationParameter(
                originalID, "AgentType", "Automatic", author=authorDN, connection=connection
            )
            if not res["OK"]:
                gLogger.error("Failed to update parent transformation agent type, now deleting", res["Message"])
                return self.deleteTransformation(transID, connection=connection)
            message = "Creation of the derived transformation (%d)" % transID
            self.__updateTransformationLogging(originalID, message, authorDN, connection=connection)
            res = self.getTransformationFiles(condDict={"TransformationID": originalID}, connection=connection)
            if not res["OK"]:
                gLogger.error("Could not get transformation files, now deleting", res["Message"])
                return self.deleteTransformation(transID, connection=connection)
            if res["Records"]:
                res = self.__insertExistingTransformationFiles(transID, res["Records"], connection=connection)
                if not res["OK"]:
                    gLogger.error("Could not insert files, now deleting", res["Message"])
                    return self.deleteTransformation(transID, connection=connection)

        # Add files to the DataFiles table
        catalog = FileCatalog()
        if addFiles and inputMetaQuery:
            res = catalog.findFilesByMetadata(inputMetaQuery)
            if not res["OK"]:
                gLogger.error("Failed to find files to be added to the transformation", res["Message"])
                return res
            filesToAdd = res["Value"]
            gLogger.notice("filesToAdd", filesToAdd)
            if filesToAdd:
                connection = self.__getConnection(connection)
                res = self.__addDataFiles(filesToAdd, connection=connection)
                if not res["OK"]:
                    return res
                lfnFileIDs = res["Value"]
                # Add the files to the transformations
                fileIDs = []
                for lfn in filesToAdd:
                    if lfn in lfnFileIDs:
                        fileIDs.append(lfnFileIDs[lfn])
                res = self.__addFilesToTransformation(transID, fileIDs, connection=connection)
                if not res["OK"]:
                    gLogger.error("Failed to add files to transformation", "%s %s" % (transID, res["Message"]))
        message = "Created transformation %d" % transID

        self.__updateTransformationLogging(transID, message, authorDN, connection=connection)
        return S_OK(transID)

    def getTransformations(
        self,
        condDict=None,
        older=None,
        newer=None,
        timeStamp="LastUpdate",
        orderAttribute=None,
        limit=None,
        extraParams=False,
        offset=None,
        connection=False,
        columns=None,
    ):
        """Get parameters of all the Transformations with support for the web standard structure"""
        connection = self.__getConnection(connection)
        if columns is None:
            columns = self.TRANSPARAMS
        else:
            columns = [c for c in columns if c in self.TRANSPARAMS]
        req = "SELECT %s FROM Transformations %s" % (
            intListToString(columns),
            self.buildCondition(condDict, older, newer, timeStamp, orderAttribute, limit, offset=offset),
        )
        res = self._query(req, connection)
        if not res["OK"]:
            return res
        if condDict is None:
            condDict = {}
        webList = []
        resultList = []
        for row in res["Value"]:
            # Prepare the structure for the web
            rList = [str(item) if not isinstance(item, six.integer_types) else item for item in row]
            transDict = dict(zip(columns, row))
            webList.append(rList)
            if extraParams:
                res = self.__getAdditionalParameters(transDict["TransformationID"], connection=connection)
                if not res["OK"]:
                    return res
                transDict.update(res["Value"])
            resultList.append(transDict)
        result = S_OK(resultList)
        result["Records"] = webList
        result["ParameterNames"] = columns
        return result

    def getTransformation(self, transName, extraParams=False, connection=False):
        """Get Transformation definition and parameters of Transformation identified by TransformationID"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        res = self.getTransformations(
            condDict={"TransformationID": transID}, extraParams=extraParams, connection=connection
        )
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR("Transformation %s did not exist" % transName)
        return S_OK(res["Value"][0])

    def getTransformationParameters(self, transName, parameters, connection=False):
        """Get the requested parameters for a supplied transformation"""
        if isinstance(parameters, six.string_types):
            parameters = [parameters]
        extraParams = bool(set(parameters) - set(self.TRANSPARAMS))
        res = self.getTransformation(transName, extraParams=extraParams, connection=connection)
        if not res["OK"]:
            return res
        transParams = res["Value"]
        paramDict = {}
        for reqParam in parameters:
            if reqParam not in transParams:
                return S_ERROR("Parameter %s not defined for transformation %s" % (reqParam, transName))
            paramDict[reqParam] = transParams[reqParam]
        if len(paramDict) == 1:
            return S_OK(paramDict[reqParam])
        return S_OK(paramDict)

    def getTransformationWithStatus(self, status, connection=False):
        """Gets a list of the transformations with the supplied status"""
        req = "SELECT TransformationID FROM Transformations WHERE Status = '%s';" % status
        res = self._query(req, conn=connection)
        if not res["OK"]:
            return res
        transIDs = [tupleIn[0] for tupleIn in res["Value"]]
        return S_OK(transIDs)

    def getTableDistinctAttributeValues(
        self, table, attributes, selectDict, older=None, newer=None, timeStamp=None, connection=False
    ):
        tableFields = {
            "Transformations": self.TRANSPARAMS,
            "TransformationTasks": self.TASKSPARAMS,
            "TransformationFiles": self.TRANSFILEPARAMS,
        }
        possibleFields = tableFields.get(table, [])
        return self.__getTableDistinctAttributeValues(
            table, possibleFields, attributes, selectDict, older, newer, timeStamp, connection=connection
        )

    def __getTableDistinctAttributeValues(
        self, table, possible, attributes, selectDict, older, newer, timeStamp, connection=False
    ):
        connection = self.__getConnection(connection)
        attributeValues = {}
        for attribute in attributes:
            if possible and (attribute not in possible):
                return S_ERROR("Requested attribute (%s) does not exist in table %s" % (attribute, table))
            res = self.getDistinctAttributeValues(
                table,
                attribute,
                condDict=selectDict,
                older=older,
                newer=newer,
                timeStamp=timeStamp,
                connection=connection,
            )
            if not res["OK"]:
                return S_ERROR("Failed to serve values for attribute %s in table %s" % (attribute, table))
            attributeValues[attribute] = res["Value"]
        return S_OK(attributeValues)

    def __updateTransformationParameter(self, transID, paramName, paramValue, connection=False):
        if paramName not in self.mutable:
            return S_ERROR("Can not update the '%s' transformation parameter" % paramName)
        if paramName == "Body":
            res = self._escapeString(paramValue)
            if not res["OK"]:
                return S_ERROR("Failed to parse parameter value")
            paramValue = res["Value"]
            req = "UPDATE Transformations SET %s=%s, LastUpdate=UTC_TIMESTAMP() WHERE TransformationID=%d" % (
                paramName,
                paramValue,
                transID,
            )
            return self._update(req, connection)
        req = "UPDATE Transformations SET %s='%s', LastUpdate=UTC_TIMESTAMP() WHERE TransformationID=%d" % (
            paramName,
            paramValue,
            transID,
        )
        return self._update(req, connection)

    def _getTransformationID(self, transName, connection=False):
        """Method returns ID of transformation with the name=<name>"""
        try:
            transName = int(transName)
            cmd = "SELECT TransformationID from Transformations WHERE TransformationID=%d;" % transName
        except ValueError:
            if not isinstance(transName, six.string_types):
                return S_ERROR("Transformation should be ID or name")
            cmd = "SELECT TransformationID from Transformations WHERE TransformationName='%s';" % transName
        res = self._query(cmd, connection)
        if not res["OK"]:
            gLogger.error(
                "Failed to obtain transformation ID for transformation", "%s: %s" % (transName, res["Message"])
            )
            return res
        elif not res["Value"]:
            gLogger.verbose("Transformation %s does not exist" % (transName))
            return S_ERROR("Transformation does not exist")
        return S_OK(res["Value"][0][0])

    def __deleteTransformation(self, transID, connection=False):
        req = "DELETE FROM Transformations WHERE TransformationID=%d;" % transID
        return self._update(req, connection)

    def __updateFilterQueries(self, connection=False):
        """Get filters for all defined input streams in all the transformations."""
        resultList = []
        res = self.getTransformations(
            condDict={"Status": ["New", "Active", "Stopped", "Flush", "Completing"]}, connection=connection
        )
        if not res["OK"]:
            return res

        transIDs = res["Value"]
        for transDict in transIDs:
            transID = str(transDict["TransformationID"])
            res = self.getTransformationMetaQuery(transID, "Input")
            if not res["OK"]:
                continue
            resultList.append((transID, res["Value"]))

        self.filterQueries = resultList
        return S_OK(resultList)

    ###########################################################################
    #
    # These methods manipulate the AdditionalParameters tables
    #
    def setTransformationParameter(self, transName, paramName, paramValue, author="", connection=False):
        """Add a parameter for the supplied transformations"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        message = ""
        if paramName in self.TRANSPARAMS:
            res = self.__updateTransformationParameter(transID, paramName, paramValue, connection=connection)
            if res["OK"]:
                pv = self._escapeString(paramValue)
                if not pv["OK"]:
                    return S_ERROR("Failed to parse parameter value")
                paramValue = pv["Value"]
                if paramName == "Body":
                    message = "Body updated"
                else:
                    message = "%s updated to %s" % (paramName, paramValue)
        else:
            res = self.__addAdditionalTransformationParameter(transID, paramName, paramValue, connection=connection)
            if res["OK"]:
                message = "Added additional parameter %s" % paramName
        if message:
            self.__updateTransformationLogging(transID, message, author, connection=connection)
        return res

    def getAdditionalParameters(self, transName, connection=False):
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        return self.__getAdditionalParameters(transID, connection=connection)

    def deleteTransformationParameter(self, transName, paramName, author="", connection=False):
        """Delete a parameter from the additional parameters table"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        if paramName in self.TRANSPARAMS:
            return S_ERROR("Can not delete core transformation parameter")
        res = self.__deleteTransformationParameters(transID, parameters=[paramName], connection=connection)
        if not res["OK"]:
            return res
        self.__updateTransformationLogging(
            transID, "Removed additional parameter %s" % paramName, author, connection=connection
        )
        return res

    def __addAdditionalTransformationParameter(self, transID, paramName, paramValue, connection=False):
        req = "DELETE FROM AdditionalParameters WHERE TransformationID=%d AND ParameterName='%s'" % (transID, paramName)
        res = self._update(req, connection)
        if not res["OK"]:
            return res
        res = self._escapeString(paramValue)
        if not res["OK"]:
            return S_ERROR("Failed to parse parameter value")
        paramValue = res["Value"]
        paramType = "StringType"
        if isinstance(paramValue, six.integer_types):
            paramType = "IntType"
        req = "INSERT INTO AdditionalParameters (%s) VALUES (%s,'%s',%s,'%s');" % (
            ", ".join(self.ADDITIONALPARAMETERS),
            transID,
            paramName,
            paramValue,
            paramType,
        )
        return self._update(req, connection)

    def __getAdditionalParameters(self, transID, connection=False):
        req = "SELECT %s FROM AdditionalParameters WHERE TransformationID = %d" % (
            ", ".join(self.ADDITIONALPARAMETERS),
            transID,
        )
        res = self._query(req, connection)
        if not res["OK"]:
            return res
        paramDict = {}
        for _transID, parameterName, parameterValue, parameterType in res["Value"]:
            if parameterType in ("IntType", "LongType"):
                parameterValue = int(parameterValue)
            paramDict[parameterName] = parameterValue
        return S_OK(paramDict)

    def __deleteTransformationParameters(self, transID, parameters=None, connection=False):
        """Remove the parameters associated to a transformation"""
        if parameters is None:
            parameters = []
        req = "DELETE FROM AdditionalParameters WHERE TransformationID=%d" % transID
        if parameters:
            req = "%s AND ParameterName IN (%s);" % (req, stringListToString(parameters))
        return self._update(req, connection)

    ###########################################################################
    #
    # These methods manipulate the TransformationFiles table
    #

    def addFilesToTransformation(self, transName, lfns, connection=False):
        """Add a list of LFNs to the transformation directly"""
        gLogger.info(
            "TransformationDB.addFilesToTransformation:"
            " Attempting to add %s files to transformations: %s" % (len(lfns), transName)
        )
        if not lfns:
            return S_ERROR("Zero length LFN list")
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        # Add missing files if necessary (__addDataFiles does the job)
        res = self.__addDataFiles(lfns, connection=connection)
        if not res["OK"]:
            return res
        fileIDs = dict((fileID, lfn) for lfn, fileID in res["Value"].items())
        # Attach files to transformation
        successful = {}
        if fileIDs:
            res = self.__addFilesToTransformation(transID, list(fileIDs), connection=connection)
            if not res["OK"]:
                return res
            for fileID in fileIDs:
                lfn = fileIDs[fileID]
                successful[lfn] = "Added" if fileID in res["Value"] else "Present"
        resDict = {"Successful": successful, "Failed": {}}
        return S_OK(resDict)

    def getTransformationFiles(
        self,
        condDict=None,
        older=None,
        newer=None,
        timeStamp="LastUpdate",
        orderAttribute=None,
        limit=None,
        offset=None,
        connection=False,
    ):
        """Get files for the supplied transformations with support for the web standard structure"""
        connection = self.__getConnection(connection)
        req = "SELECT %s FROM TransformationFiles" % (intListToString(self.TRANSFILEPARAMS))
        originalFileIDs = {}
        if condDict is None:
            condDict = {}
        if condDict or older or newer:
            lfns = condDict.pop("LFN", None)
            if lfns:
                if isinstance(lfns, six.string_types):
                    lfns = [lfns]
                res = self.__getFileIDsForLfns(lfns, connection=connection)
                if not res["OK"]:
                    return res
                originalFileIDs = res["Value"][0]
                condDict["FileID"] = list(originalFileIDs)

            for val in condDict.values():
                if not val:
                    return S_OK([])

            req = "%s %s" % (
                req,
                self.buildCondition(condDict, older, newer, timeStamp, orderAttribute, limit, offset=offset),
            )
        res = self._query(req, connection)
        if not res["OK"]:
            return res

        transFiles = res["Value"]
        fileIDs = [int(row[1]) for row in transFiles]
        webList = []
        resultList = []
        if not fileIDs:
            originalFileIDs = {}
        else:
            if not originalFileIDs:
                res = self.__getLfnsForFileIDs(fileIDs, connection=connection)
                if not res["OK"]:
                    return res
                originalFileIDs = res["Value"][1]
            for row in transFiles:
                lfn = originalFileIDs[row[1]]
                # Prepare the structure for the web
                fDict = {"LFN": lfn}
                fDict.update(dict(zip(self.TRANSFILEPARAMS, row)))
                # Note: the line below is returning "None" if the item is None... This seems to work but is ugly...
                rList = [lfn] + [str(item) if not isinstance(item, six.integer_types) else item for item in row]
                webList.append(rList)
                resultList.append(fDict)
        result = S_OK(resultList)
        result["Records"] = webList
        result["ParameterNames"] = ["LFN"] + self.TRANSFILEPARAMS
        return result

    def getFileSummary(self, lfns, connection=False):
        """Get file status summary in all the transformations"""
        connection = self.__getConnection(connection)
        condDict = {"LFN": lfns}
        res = self.getTransformationFiles(condDict=condDict, connection=connection)
        if not res["OK"]:
            return res
        resDict = {}
        for fileDict in res["Value"]:
            resDict.setdefault(fileDict["LFN"], {})[fileDict["TransformationID"]] = fileDict
        failedDict = dict.fromkeys(set(lfns) - set(resDict), "Did not exist in the Transformation database")
        return S_OK({"Successful": resDict, "Failed": failedDict})

    def setFileStatusForTransformation(self, transID, fileStatusDict=None, connection=False):
        """Set file status for the given transformation, based on
        fileStatusDict {fileID_A: ('statusA',errorA), fileID_B: ('statusB',errorB), ...}

        The ErrorCount is incremented if errorA flag is True
        """
        if not fileStatusDict:
            return S_OK()

        # Building the request with "ON DUPLICATE KEY UPDATE"
        reqBase = "INSERT INTO TransformationFiles (TransformationID, FileID, Status, ErrorCount, LastUpdate) VALUES "

        # Get fileID and status for each case: error and no error
        statusFileDict = {}
        for fileID, (status, error) in fileStatusDict.items():
            statusFileDict.setdefault(error, []).append((fileID, status))

        for error, fileIDStatusList in statusFileDict.items():
            req = reqBase + ",".join(
                "(%d, %d, '%s', 0, UTC_TIMESTAMP())" % (transID, fileID, status) for fileID, status in fileIDStatusList
            )
            if error:
                # Increment the error counter when we requested
                req += " ON DUPLICATE KEY UPDATE Status=VALUES(Status),ErrorCount=ErrorCount+1,LastUpdate=VALUES(LastUpdate)"
            else:
                req += " ON DUPLICATE KEY UPDATE Status=VALUES(Status),LastUpdate=VALUES(LastUpdate)"

            result = self._update(req, connection)
            if not result["OK"]:
                return result
        return S_OK()

    def getTransformationStats(self, transName, connection=False):
        """Get number of files in Transformation Table for each status"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        res = self.getCounters("TransformationFiles", ["TransformationID", "Status"], {"TransformationID": transID})
        if not res["OK"]:
            return res
        statusDict = dict(
            (attrDict["Status"], count) for attrDict, count in res["Value"] if "-" not in attrDict["Status"]
        )
        statusDict["Total"] = sum(statusDict.values())
        return S_OK(statusDict)

    def getTransformationFilesCount(self, transName, field, selection=None, connection=False):
        """Get the number of files in the TransformationFiles table grouped by the supplied field"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        if selection is None:
            selection = {}
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        selection["TransformationID"] = transID
        if field not in self.TRANSFILEPARAMS:
            return S_ERROR("Supplied field not in TransformationFiles table")
        res = self.getCounters("TransformationFiles", ["TransformationID", field], selection)
        if not res["OK"]:
            return res
        countDict = dict((attrDict[field], count) for attrDict, count in res["Value"])
        countDict["Total"] = sum(countDict.values())
        return S_OK(countDict)

    def __addFilesToTransformation(self, transID, fileIDs, connection=False):
        req = "SELECT FileID from TransformationFiles"
        req = req + " WHERE TransformationID = %d AND FileID IN (%s);" % (transID, intListToString(fileIDs))
        res = self._query(req, connection)
        if not res["OK"]:
            return res
        for tupleIn in res["Value"]:
            fileIDs.remove(tupleIn[0])
        if not fileIDs:
            return S_OK([])
        req = "INSERT INTO TransformationFiles (TransformationID,FileID,LastUpdate,InsertedTime) VALUES"
        for fileID in fileIDs:
            req = "%s (%d,%d,UTC_TIMESTAMP(),UTC_TIMESTAMP())," % (req, transID, fileID)
        req = req.rstrip(",")
        res = self._update(req, connection)
        if not res["OK"]:
            return res
        return S_OK(fileIDs)

    def __insertExistingTransformationFiles(self, transID, fileTuplesList, connection=False):
        """Inserting already transformation files in TransformationFiles table (e.g. for deriving transformations)"""
        gLogger.info("Inserting %d files in TransformationFiles" % len(fileTuplesList))

        # splitting in various chunks, in case it is too big
        for fileTuples in breakListIntoChunks(fileTuplesList, 10000):

            gLogger.verbose(
                "Adding first %d files in TransformationFiles (out of %d)" % (len(fileTuples), len(fileTuplesList))
            )
            req = "INSERT INTO TransformationFiles (TransformationID,Status,TaskID,FileID,TargetSE,UsedSE,LastUpdate) VALUES"
            candidates = False

            for ft in fileTuples:
                _lfn, originalID, fileID, status, taskID, targetSE, usedSE, _errorCount, _lastUpdate, _insertTime = ft[
                    :10
                ]
                if status not in ("Removed",):
                    candidates = True
                    if not re.search("-", status):
                        status = "%s-inherited" % status
                        if taskID:
                            # Should be readable up to 999,999 tasks: that field is an int(11) in the DB, not a string
                            taskID = 1000000 * int(originalID) + int(taskID)
                    req = "%s (%d,'%s','%d',%d,'%s','%s',UTC_TIMESTAMP())," % (
                        req,
                        transID,
                        status,
                        taskID,
                        fileID,
                        targetSE,
                        usedSE,
                    )
            if not candidates:
                continue

            req = req.rstrip(",")
            res = self._update(req, connection)
            if not res["OK"]:
                return res

        return S_OK()

    def __assignTransformationFile(self, transID, taskID, se, fileIDs, connection=False):
        """Make necessary updates to the TransformationFiles table for the newly created task"""
        req = "UPDATE TransformationFiles SET TaskID='%d',UsedSE='%s',Status='Assigned',LastUpdate=UTC_TIMESTAMP()"
        req = (req + " WHERE TransformationID = %d AND FileID IN (%s);") % (
            taskID,
            se,
            transID,
            intListToString(fileIDs),
        )
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("Failed to assign file to task", res["Message"])
        fileTuples = []
        for fileID in fileIDs:
            fileTuples.append(("(%d,%d,%d)" % (transID, fileID, taskID)))
        req = "INSERT INTO TransformationFileTasks (TransformationID,FileID,TaskID) VALUES %s" % ",".join(fileTuples)
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("Failed to assign file to task", res["Message"])
        return res

    def __setTransformationFileStatus(self, fileIDs, status, connection=False):
        req = "UPDATE TransformationFiles SET Status = '%s' WHERE FileID IN (%s);" % (status, intListToString(fileIDs))
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("Failed to update file status", res["Message"])
        return res

    def __setTransformationFileUsedSE(self, fileIDs, usedSE, connection=False):
        req = "UPDATE TransformationFiles SET UsedSE = '%s' WHERE FileID IN (%s);" % (usedSE, intListToString(fileIDs))
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("Failed to update file usedSE", res["Message"])
        return res

    def __resetTransformationFile(self, transID, taskID, connection=False):
        req = (
            "UPDATE TransformationFiles SET TaskID=NULL, UsedSE='Unknown', Status='Unused'\
     WHERE TransformationID = %d AND TaskID=%d;"
            % (transID, taskID)
        )
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("Failed to reset transformation file", res["Message"])
        return res

    def __deleteTransformationFiles(self, transID, connection=False):
        """Remove the files associated to a transformation.
        It also tries to remove the associated DataFiles.
        If these DataFiles are still used by other transformations, they
        will be kept thanks to the ForeignKey constraint.
        In the very unlikely event of removing a file that was juuuuuuuust about to be
        used by another transformation, well, tough luck, but the other transformation
        will succeed at the next attempt to insert the file.
        """
        # The IGNORE keyword will make sure we do not abort the full removal
        # on a foreign key error
        # https://dev.mysql.com/doc/refman/5.7/en/sql-mode.html#ignore-strict-comparison
        req = (
            "DELETE IGNORE tf, df \
           FROM TransformationFiles tf \
           JOIN DataFiles df \
           ON tf.FileID=df.FileID \
           WHERE TransformationID = %d;"
            % transID
        )
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("Failed to delete transformation files", res["Message"])
        return res

    ###########################################################################
    #
    # These methods manipulate the TransformationFileTasks table
    #

    def __deleteTransformationFileTask(self, transID, taskID, connection=False):
        """Delete the file associated to a given task of a given transformation
        from the TransformationFileTasks table for transformation with TransformationID and TaskID
        """
        req = "DELETE FROM TransformationFileTasks WHERE TransformationID=%d AND TaskID=%d" % (transID, taskID)
        return self._update(req, connection)

    def __deleteTransformationFileTasks(self, transID, connection=False):
        """Remove all associations between files, tasks and a transformation"""
        req = "DELETE FROM TransformationFileTasks WHERE TransformationID = %d;" % transID
        res = self._update(req, connection)
        if not res["OK"]:
            gLogger.error("Failed to delete transformation files/task history", res["Message"])
        return res

    ###########################################################################
    #
    # These methods manipulate the TransformationTasks table
    #

    def getTransformationTasks(
        self,
        condDict=None,
        older=None,
        newer=None,
        timeStamp="CreationTime",
        orderAttribute=None,
        limit=None,
        inputVector=False,
        offset=None,
        connection=False,
    ):
        connection = self.__getConnection(connection)
        req = "SELECT %s FROM TransformationTasks %s" % (
            intListToString(self.TASKSPARAMS),
            self.buildCondition(condDict, older, newer, timeStamp, orderAttribute, limit, offset=offset),
        )
        res = self._query(req, connection)
        if not res["OK"]:
            return res
        if condDict is None:
            condDict = {}
        webList = []
        resultList = []
        for row in res["Value"]:
            # Prepare the structure for the web
            rList = [str(item) if not isinstance(item, six.integer_types) else item for item in row]
            taskDict = dict(zip(self.TASKSPARAMS, row))
            webList.append(rList)
            if inputVector:
                taskDict["InputVector"] = ""
                taskID = taskDict["TaskID"]
                transID = taskDict["TransformationID"]
                res = self.getTaskInputVector(transID, taskID)
                if res["OK"]:
                    if taskID in res["Value"]:
                        taskDict["InputVector"] = res["Value"][taskID]
                else:
                    return res
            resultList.append(taskDict)
        result = S_OK(resultList)
        result["Records"] = webList
        result["ParameterNames"] = self.TASKSPARAMS
        return result

    def getTasksForSubmission(
        self, transName, numTasks=1, site="", statusList=None, older=None, newer=None, connection=False
    ):
        """Select tasks with the given status (and site) for submission"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        if statusList is None:
            statusList = ["Created"]
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        condDict = {"TransformationID": transID}
        if statusList:
            condDict["ExternalStatus"] = statusList
        if site:
            numTasks = 0
        res = self.getTransformationTasks(
            condDict=condDict,
            older=older,
            newer=newer,
            timeStamp="CreationTime",
            orderAttribute=None,
            limit=numTasks,
            inputVector=True,
            connection=connection,
        )
        if not res["OK"]:
            return res
        tasks = res["Value"]

        # Now prepare the tasks
        resultDict = {}

        for taskDict in tasks:
            if len(resultDict) >= numTasks:
                break
            taskDict["Status"] = taskDict.pop("ExternalStatus")
            taskDict["InputData"] = taskDict.pop("InputVector")
            taskDict.pop("LastUpdateTime")
            taskDict.pop("CreationTime")
            taskDict.pop("ExternalID")
            taskID = taskDict["TaskID"]
            resultDict[taskID] = taskDict
            if site:
                resultDict[taskID]["Site"] = site

        return S_OK(resultDict)

    def deleteTasks(self, transName, taskIDbottom, taskIDtop, author="", connection=False):
        """Delete tasks with taskID range in transformation"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        for taskID in range(taskIDbottom, taskIDtop + 1):
            res = self.__removeTransformationTask(transID, taskID, connection=connection)
            if not res["OK"]:
                return res
        message = "Deleted tasks from %d to %d" % (taskIDbottom, taskIDtop)
        self.__updateTransformationLogging(transID, message, author, connection=connection)
        return res

    def reserveTask(self, transName, taskID, connection=False):
        """Reserve the taskID from transformation for submission"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        res = self.__checkUpdate(
            "TransformationTasks",
            "ExternalStatus",
            "Reserved",
            {"TransformationID": transID, "TaskID": taskID},
            connection=connection,
        )
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR("Failed to set Reserved status for job %d - already Reserved" % int(taskID))
        # The job is reserved, update the time stamp
        res = self.setTaskStatus(transID, taskID, "Reserved", connection=connection)
        if not res["OK"]:
            return S_ERROR("Failed to set Reserved status for job %d - failed to update the time stamp" % int(taskID))
        return S_OK()

    def setTaskStatusAndWmsID(self, transName, taskID, status, taskWmsID, connection=False):
        """Set status and ExternalID for job with taskID in production with transformationID"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        # Set ID first in order to be sure there is no status set without the ID being set
        res = self.__setTaskParameterValue(transID, taskID, "ExternalID", taskWmsID, connection=connection)
        if not res["OK"]:
            return res
        return self.__setTaskParameterValue(transID, taskID, "ExternalStatus", status, connection=connection)

    def setTaskStatus(self, transName, taskID, status, connection=False):
        """Set status for job with taskID in production with transformationID"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        if not isinstance(taskID, list):
            taskIDList = [taskID]
        else:
            taskIDList = list(taskID)
        for taskID in taskIDList:
            res = self.__setTaskParameterValue(transID, taskID, "ExternalStatus", status, connection=connection)
            if not res["OK"]:
                return res
        return S_OK()

    def getTransformationTaskStats(self, transName="", connection=False):
        """Returns dictionary with number of jobs per status for the given production."""
        connection = self.__getConnection(connection)
        if transName:
            res = self._getTransformationID(transName, connection=connection)
            if not res["OK"]:
                gLogger.error("Failed to get ID for transformation", res["Message"])
                return res
            res = self.getCounters(
                "TransformationTasks", ["ExternalStatus"], {"TransformationID": res["Value"]}, connection=connection
            )
        else:
            res = self.getCounters(
                "TransformationTasks", ["ExternalStatus", "TransformationID"], {}, connection=connection
            )
        if not res["OK"]:
            return res
        statusDict = {}
        total = 0
        for attrDict, count in res["Value"]:
            status = attrDict["ExternalStatus"]
            statusDict[status] = count
            total += count
        statusDict["TotalCreated"] = total
        return S_OK(statusDict)

    def __setTaskParameterValue(self, transID, taskID, paramName, paramValue, connection=False):
        req = "UPDATE TransformationTasks SET %s='%s', LastUpdateTime=UTC_TIMESTAMP()" % (paramName, paramValue)
        req = req + " WHERE TransformationID=%d AND TaskID=%d;" % (transID, taskID)
        return self._update(req, connection)

    def __deleteTransformationTasks(self, transID, connection=False):
        """Delete all the tasks from the TransformationTasks table for transformation with TransformationID"""
        req = "DELETE FROM TransformationTasks WHERE TransformationID=%d" % transID
        return self._update(req, connection)

    def __deleteTransformationTask(self, transID, taskID, connection=False):
        """Delete the task from the TransformationTasks table for transformation with TransformationID"""
        req = "DELETE FROM TransformationTasks WHERE TransformationID=%d AND TaskID=%d" % (transID, taskID)
        return self._update(req, connection)

    def __deleteTransformationMetaQueries(self, transID, connection=False):
        """Delete all the meta queries from the TransformationMetaQueries table for transformation with TransformationID"""
        req = "DELETE FROM TransformationMetaQueries WHERE TransformationID=%d" % transID
        return self._update(req, connection)

    ####################################################################
    #
    # These methods manipulate the TransformationMetaQueries table. It replaces all methods used to manipulate
    # the old InputDataQuery table
    #

    def createTransformationMetaQuery(self, transName, queryDict, queryType, author="", connection=False):
        """Add a Meta Query to a given transformation"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        return self.__addMetaQuery(transID, queryDict, queryType, author=author, connection=connection)

    def __addMetaQuery(self, transID, queryDict, queryType, author="", connection=False):
        """Insert the Meta Query into the TransformationMetaQuery table"""
        res = self.getTransformationMetaQuery(transID, queryType, connection=connection)
        if res["OK"]:
            return S_ERROR("Meta query already exists for transformation")
        if not cmpError(res, ENOENT):
            return res

        for parameterName in sorted(queryDict):
            parameterValue = queryDict[parameterName]
            if not parameterValue:
                continue
            parameterType = "String"
            if isinstance(parameterValue, (list, tuple)):
                if isinstance(parameterValue[0], six.integer_types):
                    parameterType = "Integer"
                    parameterValue = [str(x) for x in parameterValue]
                parameterValue = ";;;".join(parameterValue)
            else:
                if isinstance(parameterValue, six.integer_types):
                    parameterType = "Integer"
                    parameterValue = str(parameterValue)
                if isinstance(parameterValue, dict):
                    parameterType = "Dict"
                    parameterValue = str(parameterValue)

            res = self.insertFields(
                "TransformationMetaQueries",
                ["TransformationID", "MetaDataName", "MetaDataValue", "MetaDataType", "QueryType"],
                [transID, parameterName, parameterValue, parameterType, queryType],
                conn=connection,
            )
            if not res["OK"]:
                message = "Failed to add meta query"
                self.deleteTransformationMetaQuery(transID, queryType, connection=connection)
                break
            else:
                message = "Added meta data query"

        self.__updateTransformationLogging(transID, message, author, connection=connection)
        return res

    def deleteTransformationMetaQuery(self, transName, queryType, author="", connection=False):
        """Remove a Meta Query from the TransformationMetaQueries table"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        res = self._escapeString(queryType)
        if not res["OK"]:
            return S_ERROR("Failed to parse the transformation query type")
        queryType = res["Value"]
        req = "DELETE FROM TransformationMetaQueries WHERE TransformationID=%d AND QueryType=%s;" % (transID, queryType)
        res = self._update(req, connection)
        if not res["OK"]:
            return res
        if res["Value"]:
            # Add information to the transformation logging
            message = "Deleted meta data query"
            self.__updateTransformationLogging(transID, message, author, connection=connection)
        return res

    def getTransformationMetaQuery(self, transName, queryType, connection=False):
        """Get the Meta Query for a given transformation.

        :param transName: transformation name or ID
        :type transName: str or int
        :param str queryType: 'Input' or 'Output' query
        :param connection: DB connection
        :returns: S_OK with query dictionary, S_ERROR, ENOENT if no query defined
        """
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        res = self._escapeString(queryType)
        if not res["OK"]:
            return S_ERROR("Failed to parse the transformation query type")
        queryType = res["Value"]
        req = "SELECT MetaDataName,MetaDataValue,MetaDataType FROM TransformationMetaQueries"
        req = req + " WHERE TransformationID=%d AND QueryType=%s;" % (transID, queryType)
        res = self._query(req, connection)
        if not res["OK"]:
            return res
        queryDict = {}
        for parameterName, parameterValue, parameterType in res["Value"]:
            if re.search(";;;", str(parameterValue)):
                parameterValue = parameterValue.split(";;;")
                if parameterType == "Integer":
                    parameterValue = [int(x) for x in parameterValue]
            elif parameterType == "Integer":
                parameterValue = int(parameterValue)
            elif parameterType == "Dict":
                parameterValue = eval(parameterValue)
            queryDict[parameterName] = parameterValue
        if not queryDict:
            return S_ERROR(ENOENT, "No MetaQuery found for transformation")
        return S_OK(queryDict)

    ###########################################################################
    #
    # These methods manipulate the TaskInputs table
    #

    def getTaskInputVector(self, transName, taskID, connection=False):
        """Get input vector for the given task"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        if not isinstance(taskID, list):
            taskIDList = [taskID]
        else:
            taskIDList = list(taskID)
        taskString = ",".join(["'%s'" % x for x in taskIDList])
        req = "SELECT TaskID,InputVector FROM TaskInputs WHERE TaskID in (%s) AND TransformationID='%d';" % (
            taskString,
            transID,
        )
        res = self._query(req)
        inputVectorDict = {}
        if not res["OK"]:
            return res
        elif res["Value"]:
            for row in res["Value"]:
                inputVectorDict[row[0]] = row[1]
        return S_OK(inputVectorDict)

    def __insertTaskInputs(self, transID, taskID, lfns, connection=False):
        vector = str.join(";", lfns)
        fields = ["TransformationID", "TaskID", "InputVector"]
        values = [transID, taskID, vector]
        res = self.insertFields("TaskInputs", fields, values, connection)
        if not res["OK"]:
            gLogger.error("Failed to add input vector to task %d" % taskID)
        return res

    def __deleteTransformationTaskInputs(self, transID, taskID=0, connection=False):
        """Delete all the tasks inputs from the TaskInputs table for transformation with TransformationID"""
        req = "DELETE FROM TaskInputs WHERE TransformationID=%d" % transID
        if taskID:
            req = "%s AND TaskID=%d" % (req, int(taskID))
        return self._update(req, connection)

    ###########################################################################
    #
    # These methods manipulate the TransformationLog table
    #

    def __updateTransformationLogging(self, transName, message, authorDN, connection=False):
        """Update the Transformation log table with any modifications"""
        if not authorDN:
            res = getProxyInfo(False, False)
            if res["OK"]:
                authorDN = res["Value"]["subject"]
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        req = "INSERT INTO TransformationLog (TransformationID,Message,Author,MessageDate)"
        req = req + " VALUES (%s,'%s','%s',UTC_TIMESTAMP());" % (transID, message, authorDN)
        return self._update(req, connection)

    def getTransformationLogging(self, transName, connection=False):
        """Get logging info from the TransformationLog table"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        req = "SELECT TransformationID, Message, Author, MessageDate FROM TransformationLog"
        req = req + " WHERE TransformationID=%s ORDER BY MessageDate;" % (transID)
        res = self._query(req)
        if not res["OK"]:
            return res
        transList = []
        for transID, message, authorDN, messageDate in res["Value"]:
            transDict = {}
            transDict["TransformationID"] = transID
            transDict["Message"] = message
            transDict["AuthorDN"] = authorDN
            transDict["MessageDate"] = messageDate
            transList.append(transDict)
        return S_OK(transList)

    def __deleteTransformationLog(self, transID, connection=False):
        """Remove the entries in the transformation log for a transformation"""
        req = "DELETE FROM TransformationLog WHERE TransformationID=%d;" % transID
        return self._update(req, connection)

    ###########################################################################
    #
    # These methods manipulate the DataFiles table
    #
    def __getAllFileIDs(self, connection=False):
        """Get all the fileIDs for the supplied list of lfns"""
        req = "SELECT LFN,FileID FROM DataFiles;"
        res = self._query(req, connection)
        if not res["OK"]:
            return res
        fids = {}
        lfns = {}
        for lfn, fileID in res["Value"]:
            fids[fileID] = lfn
            lfns[lfn] = fileID
        return S_OK((fids, lfns))

    def __getFileIDsForLfns(self, lfns, connection=False):
        """Get file IDs for the given list of lfns
        warning: if the file is not present, we'll see no errors
        """
        req = "SELECT LFN,FileID FROM DataFiles WHERE LFN in (%s);" % (stringListToString(lfns))
        res = self._query(req, connection)
        if not res["OK"]:
            return res
        lfns = dict(res["Value"])
        # Reverse dictionary
        fids = dict((fileID, lfn) for lfn, fileID in lfns.items())
        return S_OK((fids, lfns))

    def __getLfnsForFileIDs(self, fileIDs, connection=False):
        """Get lfns for the given list of fileIDs"""
        req = "SELECT LFN,FileID FROM DataFiles WHERE FileID in (%s);" % stringListToString(fileIDs)
        res = self._query(req, connection)
        if not res["OK"]:
            return res
        fids = dict(res["Value"])
        # Reverse dictionary
        lfns = dict((fileID, lfn) for lfn, fileID in fids.items())
        return S_OK((fids, lfns))

    def __addDataFiles(self, lfns, connection=False):
        """Add a file to the DataFiles table and retrieve the FileIDs"""
        res = self.__getFileIDsForLfns(lfns, connection=connection)
        if not res["OK"]:
            return res
        # Insert only files not found, and assume the LFN is unique in the table
        lfnFileIDs = res["Value"][1]
        for lfn in set(lfns) - set(lfnFileIDs):
            req = "INSERT INTO DataFiles (LFN,Status) VALUES ('%s','New');" % lfn
            res = self._update(req, connection)
            # If the LFN is duplicate we get an error and ignore it
            if res["OK"]:
                lfnFileIDs[lfn] = res["lastRowId"]
        return S_OK(lfnFileIDs)

    def __setDataFileStatus(self, fileIDs, status, connection=False):
        """Set the status of the supplied files"""
        req = "UPDATE DataFiles SET Status = '%s' WHERE FileID IN (%s);" % (status, intListToString(fileIDs))
        return self._update(req, connection)

    ###########################################################################
    #
    # These methods manipulate multiple tables
    #

    def addTaskForTransformation(self, transID, lfns=None, se="Unknown", connection=False):
        """Create a new task with the supplied files for a transformation."""
        res = self._getConnectionTransID(connection, transID)
        if not res["OK"]:
            return res
        if lfns is None:
            lfns = []
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        # Be sure the all the supplied LFNs are known to the database for the supplied transformation
        fileIDs = []
        if lfns:
            res = self.getTransformationFiles(
                condDict={"TransformationID": transID, "LFN": lfns}, connection=connection
            )
            if not res["OK"]:
                return res
            foundLfns = set()
            for fileDict in res["Value"]:
                fileIDs.append(fileDict["FileID"])
                lfn = fileDict["LFN"]
                if fileDict["Status"] in self.allowedStatusForTasks:
                    foundLfns.add(lfn)
                else:
                    gLogger.error(
                        "Supplied file not in %s status but %s" % (self.allowedStatusForTasks, fileDict["Status"]), lfn
                    )
            unavailableLfns = set(lfns) - foundLfns
            if unavailableLfns:
                gLogger.error("Supplied files not found for transformation", sorted(unavailableLfns))
                return S_ERROR("Not all supplied files available in the transformation database")

        # Insert the task into the jobs table and retrieve the taskID
        self.lock.acquire()
        req = "INSERT INTO TransformationTasks(TransformationID, ExternalStatus, ExternalID, TargetSE,"
        req = req + " CreationTime, LastUpdateTime)"
        req = req + " VALUES (%s,'%s','%d','%s', UTC_TIMESTAMP(), UTC_TIMESTAMP());" % (transID, "Created", 0, se)
        res = self._update(req, connection)
        if not res["OK"]:
            self.lock.release()
            gLogger.error("Failed to publish task for transformation", res["Message"])
            return res

        # With InnoDB, TaskID is computed by a trigger, which sets the local variable @last (per connection)
        # @last is the last insert TaskID. With multi-row inserts, will be the first new TaskID inserted.
        # The trigger TaskID_Generator must be present with the InnoDB schema (defined in TransformationDB.sql)
        if self.isTransformationTasksInnoDB:
            res = self._query("SELECT @last;", connection)
        else:
            res = self._query("SELECT LAST_INSERT_ID();", connection)

        self.lock.release()
        if not res["OK"]:
            return res
        taskID = int(res["Value"][0][0])
        gLogger.verbose("Published task %d for transformation %d." % (taskID, transID))
        # If we have input data then update their status, and taskID in the transformation table
        if lfns:
            res = self.__insertTaskInputs(transID, taskID, lfns, connection=connection)
            if not res["OK"]:
                self.__removeTransformationTask(transID, taskID, connection=connection)
                return res
            res = self.__assignTransformationFile(transID, taskID, se, fileIDs, connection=connection)
            if not res["OK"]:
                self.__removeTransformationTask(transID, taskID, connection=connection)
                return res
        return S_OK(taskID)

    def extendTransformation(self, transName, nTasks, author="", connection=False):
        """Extend SIMULATION type transformation by nTasks number of tasks"""
        connection = self.__getConnection(connection)
        res = self.getTransformation(transName, connection=connection)
        if not res["OK"]:
            gLogger.error("Failed to get transformation details", res["Message"])
            return res
        transType = res["Value"]["Type"]
        transID = res["Value"]["TransformationID"]
        extendableProds = Operations().getValue("Transformations/ExtendableTransfTypes", ["Simulation", "MCSimulation"])
        if transType.lower() not in [ep.lower() for ep in extendableProds]:
            return S_ERROR("Can not extend non-SIMULATION type production")
        taskIDs = []
        for _task in range(nTasks):
            res = self.addTaskForTransformation(transID, connection=connection)
            if not res["OK"]:
                return res
            taskIDs.append(res["Value"])
        # Add information to the transformation logging
        message = "Transformation extended by %d tasks" % nTasks
        self.__updateTransformationLogging(transName, message, author, connection=connection)
        return S_OK(taskIDs)

    def cleanTransformation(self, transName, author="", connection=False):
        """Clean the transformation specified by name or id"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        res = self.__deleteTransformationFileTasks(transID, connection=connection)
        if not res["OK"]:
            return res
        res = self.__deleteTransformationFiles(transID, connection=connection)
        if not res["OK"]:
            return res
        res = self.__deleteTransformationTaskInputs(transID, connection=connection)
        if not res["OK"]:
            return res
        res = self.__deleteTransformationTasks(transID, connection=connection)
        if not res["OK"]:
            return res
        res = self.__deleteTransformationMetaQueries(transID, connection=connection)
        if not res["OK"]:
            return res

        self.__updateTransformationLogging(transID, "Transformation Cleaned", author, connection=connection)

        return S_OK(transID)

    def deleteTransformation(self, transName, author="", connection=False):
        """Remove the transformation specified by name or id"""
        res = self._getConnectionTransID(connection, transName)
        if not res["OK"]:
            return res
        connection = res["Value"]["Connection"]
        transID = res["Value"]["TransformationID"]
        res = self.cleanTransformation(transID, author=author, connection=connection)
        if not res["OK"]:
            return res
        res = self.__deleteTransformationLog(transID, connection=connection)
        if not res["OK"]:
            return res
        res = self.__deleteTransformationParameters(transID, connection=connection)
        if not res["OK"]:
            return res
        res = self.__deleteTransformation(transID, connection=connection)
        if not res["OK"]:
            return res
        res = self.__updateFilterQueries(connection=connection)
        if not res["OK"]:
            return res
        return S_OK()

    def __removeTransformationTask(self, transID, taskID, connection=False):
        res = self.__deleteTransformationTaskInputs(transID, taskID, connection=connection)
        if not res["OK"]:
            return res
        res = self.__deleteTransformationFileTask(transID, taskID, connection=connection)
        if not res["OK"]:
            return res
        res = self.__resetTransformationFile(transID, taskID, connection=connection)
        if not res["OK"]:
            return res
        return self.__deleteTransformationTask(transID, taskID, connection=connection)

    def __checkUpdate(self, table, param, paramValue, selectDict=None, connection=False):
        """Check whether the update will perform an update"""
        req = "UPDATE %s SET %s = '%s'" % (table, param, paramValue)
        if selectDict:
            req = "%s %s" % (req, self.buildCondition(selectDict))
        return self._update(req, connection)

    def __getConnection(self, connection):
        if connection:
            return connection
        res = self._getConnection()
        if res["OK"]:
            return res["Value"]
        gLogger.warn("Failed to get MySQL connection", res["Message"])
        return connection

    def _getConnectionTransID(self, connection, transName):
        connection = self.__getConnection(connection)
        res = self._getTransformationID(transName, connection=connection)
        if not res["OK"]:
            gLogger.error("Failed to get ID for transformation", res["Message"])
            return res
        transID = res["Value"]
        resDict = {"Connection": connection, "TransformationID": transID}
        return S_OK(resDict)

    ####################################################################################
    #
    #  This part should correspond to the DIRAC Standard File Catalog interface
    #
    ####################################################################################

    def exists(self, lfns, connection=False):
        """Check the presence of the lfn in the TransformationDB DataFiles table"""
        gLogger.info("TransformationDB.exists: Attempting to determine existence of %s files." % len(lfns))
        res = self.__getFileIDsForLfns(lfns, connection=connection)
        if not res["OK"]:
            return res
        fileIDs = res["Value"][0]
        failed = {}
        successful = {}
        fileIDsValues = set(fileIDs.values())
        for lfn in lfns:
            successful[lfn] = lfn in fileIDsValues
        resDict = {"Successful": successful, "Failed": failed}
        return S_OK(resDict)

    def addFile(self, fileDicts, force=False, connection=False):
        """Add the supplied lfn to the Transformations and to the DataFiles table if it passes the filter"""
        gLogger.info("TransformationDB.addFile: Attempting to add %s files." % len(fileDicts))
        successful = {}
        failed = {}
        # Determine which files pass the filters and are to be added to transformations
        transFiles = {}
        filesToAdd = []
        catalog = FileCatalog()

        for lfn in fileDicts:
            gLogger.info("addFile: Attempting to add file %s" % lfn)
            res = catalog.getFileUserMetadata(lfn)
            if not res["OK"]:
                gLogger.error("Failed to getFileUserMetadata for file", "%s: %s" % (lfn, res["Message"]))
                failed[lfn] = res["Message"]
                continue
            else:
                metadatadict = res["Value"]
            gLogger.info("Filter file with metadata", metadatadict)
            transIDs = self._filterFileByMetadata(metadatadict)
            gLogger.info("Transformations passing the filter: %s" % transIDs)
            if not (transIDs or force):  # not clear how force should be used for
                successful[lfn] = False  # True -> False bug fix: otherwise it is set to True even if transIDs is empty.
            else:
                filesToAdd.append(lfn)
                for trans in transIDs:
                    if trans not in transFiles:
                        transFiles[trans] = []
                    transFiles[trans].append(lfn)

            # Add the files to the transformations
            gLogger.info("Files to add to transformations:", filesToAdd)
            if filesToAdd:
                for transID, lfns in transFiles.items():
                    res = self.addFilesToTransformation(transID, lfns)
                    if not res["OK"]:
                        gLogger.error("Failed to add files to transformation", "%s %s" % (transID, res["Message"]))
                        return res
                    else:
                        for lfn in lfns:
                            successful[lfn] = True

        res = S_OK({"Successful": successful, "Failed": failed})
        return res

    def removeFile(self, lfns, connection=False):
        """Remove file specified by lfn from the ProcessingDB"""
        gLogger.info("TransformationDB.removeFile: Attempting to remove %s files." % len(lfns))
        failed = {}
        successful = {}
        connection = self.__getConnection(connection)
        if not lfns:
            return S_ERROR("No LFNs supplied")
        res = self.__getFileIDsForLfns(lfns, connection=connection)
        if not res["OK"]:
            return res
        fileIDs, lfnFilesIDs = res["Value"]
        for lfn in lfns:
            if lfn not in lfnFilesIDs:
                successful[lfn] = "File does not exist"
        if fileIDs:
            res = self.__setTransformationFileStatus(list(fileIDs), "Deleted", connection=connection)
            if not res["OK"]:
                return res
            res = self.__setDataFileStatus(list(fileIDs), "Deleted", connection=connection)
            if not res["OK"]:
                return S_ERROR("TransformationDB.removeFile: Failed to remove files.")
        for lfn in lfnFilesIDs:
            if lfn not in failed:
                successful[lfn] = True
        resDict = {"Successful": successful, "Failed": failed}
        return S_OK(resDict)

    def addDirectory(self, path, force=False):
        """Adds all the files stored in a given directory in file catalog"""
        gLogger.info("TransformationDB.addDirectory: Attempting to populate %s." % path)
        res = pythonCall(30, self.__addDirectory, path, force)
        if not res["OK"]:
            gLogger.error("Failed to invoke addDirectory with shifter proxy")
            return res
        return res["Value"]

    def __addDirectory(self, path, force):
        res = setupShifterProxyInEnv("ProductionManager")
        if not res["OK"]:
            return S_OK("Failed to setup shifter proxy")
        catalog = FileCatalog()
        start = time.time()
        res = catalog.listDirectory(path)
        if not res["OK"]:
            gLogger.error("TransformationDB.addDirectory: Failed to get files. %s" % res["Message"])
            return res
        if path not in res["Value"]["Successful"]:
            gLogger.error("TransformationDB.addDirectory: Failed to get files.")
            return res
        gLogger.info("TransformationDB.addDirectory: Obtained %s files in %s seconds." % (path, time.time() - start))
        successful = []
        failed = []
        for lfn in res["Value"]["Successful"][path]["Files"]:
            res = self.addFile({lfn: {}}, force=force)
            if not res["OK"] or lfn not in res["Value"]["Successful"]:
                failed.append(lfn)
            else:
                successful.append(lfn)
        return {"OK": True, "Value": len(res["Value"]["Successful"]), "Successful": successful, "Failed": failed}

    def setMetadata(self, path, usermetadatadict):
        """
        It can be applied to a file or to a directory (path).
        For a file, add the file to Transformations if the updated metadata dictionary passes the filter.
        For a directory, add the files contained in the directory to the Transformations
        if the the updated metadata dictionary passes the filter.
        """
        gLogger.info("setMetadata: Attempting to set metadata %s to: %s" % (usermetadatadict, path))
        transFiles = {}
        filesToAdd = []

        catalog = FileCatalog()
        res = catalog.isFile(path)
        if res["OK"]:
            isFile = res["Value"]["Successful"][path]
        else:
            gLogger.error("Failed isFile %s: %s" % (path, res["Message"]))
            return res
        res = catalog.isDirectory(path)
        if res["OK"]:
            isDirectory = res["Value"]["Successful"][path]
        else:
            gLogger.error("Failed isDirectory %s: %s" % (path, res["Message"]))
            return res

        if isFile:
            res = catalog.getFileUserMetadata(path)
        elif isDirectory:
            res = catalog.getDirectoryUserMetadata(path)

        if not res["OK"]:
            gLogger.error("Failed to get User Metadata %s: %s" % (path, res["Message"]))
            return res
        else:
            metadatadict = res["Value"]
        metadatadict.update(usermetadatadict)
        gLogger.info("Filter file with metadata:", metadatadict)
        transIDs = self._filterFileByMetadata(metadatadict)
        gLogger.info("Transformations passing the filter: %s" % transIDs)
        if not transIDs:
            return S_OK()
        elif isFile:
            filesToAdd.append(path)
        elif isDirectory:
            res = catalog.findFilesByMetadata(metadatadict, path)
            if not res["OK"]:
                gLogger.error("Failed to findFilesByMetadata %s: %s" % (path, res["Message"]))
                return res
            filesToAdd.extend(res["Value"])
        for trans in transIDs:
            if trans not in transFiles:
                transFiles[trans] = []
            transFiles[trans].extend(filesToAdd)

        # Add the files to the transformations
        gLogger.info("Files to add to transformations:", filesToAdd)
        if filesToAdd:
            for transID, lfns in transFiles.items():
                res = self.addFilesToTransformation(transID, lfns)
                if not res["OK"]:
                    gLogger.error("Failed to add files to transformation", "%s %s" % (transID, res["Message"]))
                    return res

        return S_OK()

    def _filterFileByMetadata(self, metadatadict):
        """Pass the input metadatadict through those currently active"""
        transIDs = []
        queries = self.filterQueries

        catalog = FileCatalog()
        gLogger.info("Filter file by queries", queries)
        res = catalog.getMetadataFields()

        if not res["OK"]:
            gLogger.error("Error in getMetadataFields: %s" % res["Message"])
            return res
        if not res["Value"]:
            gLogger.error("Error: no metadata fields defined")
            return res
        typeDict = res["Value"]["FileMetaFields"]
        typeDict.update(res["Value"]["DirectoryMetaFields"])

        for transID, query in queries:
            gLogger.info("Check the transformation status")
            res = self.getTransformationParameters(transID, "Status")
            if res["Value"] not in ["New", "Active", "Stopped", "Completing", "Flush"]:
                continue
            mq = MetaQuery(query, typeDict)
            gLogger.info("Apply query %s to metadata %s" % (mq.getMetaQuery(), metadatadict))
            res = mq.applyQuery(metadatadict)
            if not res["OK"]:
                gLogger.error("Error in applying query: %s" % res["Message"])
                return res
            elif res["Value"]:
                gLogger.info("Apply query result is True")
                transIDs.append(transID)
            else:
                gLogger.info("Apply query result is False")

        return transIDs



""" ProfileManager manages web user profiles interfacin to UserProfileDB
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import six

from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC import S_OK
from DIRAC.FrameworkSystem.DB.UserProfileDB import UserProfileDB
from DIRAC.Core.Security import Properties


class UserProfileManagerHandler(RequestHandler):
    @classmethod
    def initializeHandler(cls, serviceInfo):
        """Handler initialization"""
        cls.upDB = UserProfileDB()
        return S_OK()

    types_retrieveProfileVar = [six.string_types, six.string_types]

    def export_retrieveProfileVar(self, profileName, varName):
        """Get profile data for web"""
        credDict = self.getRemoteCredentials()
        userName = credDict["username"]
        userGroup = credDict["group"]
        return self.upDB.retrieveVar(userName, userGroup, userName, userGroup, profileName, varName)

    types_retrieveProfileVarFromUser = [six.string_types, six.string_types, six.string_types, six.string_types]

    def export_retrieveProfileVarFromUser(self, ownerName, ownerGroup, profileName, varName):
        """Get profile data for web for any user according to perms"""
        credDict = self.getRemoteCredentials()
        userName = credDict["username"]
        userGroup = credDict["group"]
        return self.upDB.retrieveVar(userName, userGroup, ownerName, ownerGroup, profileName, varName)

    types_retrieveProfileAllVars = [six.string_types]

    def export_retrieveProfileAllVars(self, profileName):
        """Get profile data for web"""
        credDict = self.getRemoteCredentials()
        userName = credDict["username"]
        userGroup = credDict["group"]
        return self.upDB.retrieveAllUserVars(userName, userGroup, profileName)

    types_storeProfileVar = [six.string_types, six.string_types, six.string_types, dict]

    def export_storeProfileVar(self, profileName, varName, data, perms):
        """Set profile data for web"""
        credDict = self.getRemoteCredentials()
        userName = credDict["username"]
        userGroup = credDict["group"]
        return self.upDB.storeVar(userName, userGroup, profileName, varName, data, perms)

    types_deleteProfileVar = [six.string_types, six.string_types]

    def export_deleteProfileVar(self, profileName, varName):
        """Set profile data for web"""
        credDict = self.getRemoteCredentials()
        userName = credDict["username"]
        userGroup = credDict["group"]
        return self.upDB.deleteVar(userName, userGroup, profileName, varName)

    types_listStatesForWeb = [dict]

    def export_listStatesForWeb(self, permission):
        retVal = self.export_getUserProfileNames(permission)
        if not retVal["OK"]:
            return retVal
        data = retVal["Value"]

        records = []
        for i in data:
            application = i.replace("Web/application/", "")
            retVal = self.export_listAvailableProfileVars(i)
            if not retVal["OK"]:
                return retVal
            states = retVal["Value"]
            for state in states:
                record = dict(zip(["user", "group", "vo", "name"], state))
                record["app"] = application
                retVal = self.export_getProfileVarPermissions(i, record["name"])
                if not retVal["OK"]:
                    return retVal
                record["permissions"] = retVal["Value"]
                records += [record]

        return S_OK(records)

    types_listAvailableProfileVars = [six.string_types]

    def export_listAvailableProfileVars(self, profileName, filterDict={}):
        """Set profile data for web"""
        credDict = self.getRemoteCredentials()
        userName = credDict["username"]
        userGroup = credDict["group"]
        return self.upDB.listVars(userName, userGroup, profileName, filterDict)

    types_getUserProfiles = []

    def export_getUserProfiles(self):
        """Get all profiles for a user"""
        credDict = self.getRemoteCredentials()
        userName = credDict["username"]
        userGroup = credDict["group"]
        return self.upDB.retrieveUserProfiles(userName, userGroup)

    types_setProfileVarPermissions = [six.string_types, six.string_types, dict]

    def export_setProfileVarPermissions(self, profileName, varName, perms):
        """Set profile data for web"""
        credDict = self.getRemoteCredentials()
        userName = credDict["username"]
        userGroup = credDict["group"]
        return self.upDB.setUserVarPerms(userName, userGroup, profileName, varName, perms)

    types_getProfileVarPermissions = [six.string_types, six.string_types]

    def export_getProfileVarPermissions(self, profileName, varName):
        """Set profile data for web"""
        credDict = self.getRemoteCredentials()
        userName = credDict["username"]
        userGroup = credDict["group"]
        return self.upDB.retrieveVarPerms(userName, userGroup, userName, userGroup, profileName, varName)

    types_deleteProfiles = [list]

    def export_deleteProfiles(self, userList):
        """
        Delete profiles for a list of users
        """
        credDict = self.getRemoteCredentials()
        requesterUserName = credDict["username"]
        if Properties.SERVICE_ADMINISTRATOR in credDict["properties"]:
            admin = True
        else:
            admin = False
        for entry in userList:
            userName = entry
            if admin or userName == requesterUserName:
                result = self.upDB.deleteUserProfile(userName)
                if not result["OK"]:
                    return result
        return S_OK()

    types_getUserProfileNames = [dict]

    def export_getUserProfileNames(self, permission):
        """
        it returns the available profile names by not taking account the permission: ReadAccess and PublishAccess
        """
        return self.upDB.getUserProfileNames(permission)



""" Test_RSS_Policy_PilotEfficiencyPolicy
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
import DIRAC.ResourceStatusSystem.Policy.PilotEfficiencyPolicy as moduleTested

################################################################################


class PilotEfficiencyPolicy_TestCase(unittest.TestCase):
    def setUp(self):
        """
        Setup
        """

        self.moduleTested = moduleTested
        self.testClass = self.moduleTested.PilotEfficiencyPolicy

    def tearDown(self):
        """
        Tear down
        """

        del self.moduleTested
        del self.testClass


################################################################################


class PilotEfficiencyPolicy_Success(PilotEfficiencyPolicy_TestCase):
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class"""

        module = self.testClass()
        self.assertEqual("PilotEfficiencyPolicy", module.__class__.__name__)

    def test_evaluate(self):
        """tests the method _evaluate"""

        module = self.testClass()

        res = module._evaluate({"OK": False, "Message": "Bo!"})
        self.assertTrue(res["OK"])
        self.assertEqual("Error", res["Value"]["Status"])
        self.assertEqual("Bo!", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": None})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("No values to take a decision", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": []})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("No values to take a decision", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("No values to take a decision", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": [{"Aborted": 0, "Deleted": 0, "Done": 0, "Failed": 0}]})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("Not enough pilots to take a decision", res["Value"]["Reason"])

        # Pilot efficiency is now available directly from the command result, in percent.
        # The key is 'PilotJobEff'  It is calculated as (total - aborted-failed)/total * 100

        result = {"Aborted": 10, "Deleted": 0, "Done": 10, "Failed": 0}
        result["PilotJobEff"] = self._pilotEff(result)

        res = module._evaluate({"OK": True, "Value": [result]})
        self.assertTrue(res["OK"])
        self.assertEqual("Banned", res["Value"]["Status"])
        self.assertEqual("Pilots Efficiency of 50.00", res["Value"]["Reason"])

        result = {"Aborted": 0, "Deleted": 0, "Done": 30, "Failed": 10}
        result["PilotJobEff"] = self._pilotEff(result)

        res = module._evaluate({"OK": True, "Value": [result]})
        self.assertTrue(res["OK"])
        self.assertEqual("Degraded", res["Value"]["Status"])
        self.assertEqual("Pilots Efficiency of 75.00", res["Value"]["Reason"])

        result = {"Aborted": 0, "Deleted": 0, "Done": 19, "Failed": 1}
        result["PilotJobEff"] = self._pilotEff(result)

        res = module._evaluate({"OK": True, "Value": [result]})
        self.assertTrue(res["OK"])
        self.assertEqual("Active", res["Value"]["Status"])
        self.assertEqual("Pilots Efficiency of 95.00", res["Value"]["Reason"])

    def _pilotEff(self, result):
        """
        Calculate pilot efficiency.

        :param result: the original dictionary of the form {status: count, status: count...}
        :return: pilot success rate in percent
        """
        return (sum(result.values()) - result["Aborted"] - result["Failed"]) / sum(result.values()) * 100


################################################################################


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(PilotEfficiencyPolicy_TestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PilotEfficiencyPolicy_Success))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



#! /usr/bin/env python
########################################################################
# File :    dirac-admin-ce-info
# Author :  Vladimir Romanovsky
########################################################################
"""
Retrieve Site Associated to a given CE

Example:
  $ dirac-admin-ce-info LCG.IN2P3.fr
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC import gConfig, gLogger, exit as Dexit
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("CE:  Name of the CE")
    Script.parseCommandLine(ignoreErrors=True)
    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    ce = Script.getPositionalArgs(group=True)

    from DIRAC.ConfigurationSystem.Client.Helpers import cfgPath
    from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getCESiteMapping

    res = getCESiteMapping(ce)
    if not res["OK"]:
        gLogger.error(res["Message"])
        Dexit(1)
    site = res["Value"][ce]

    res = gConfig.getOptionsDict(cfgPath("Resources", "Sites", site.split(".")[0], site, "CEs", ce))
    if not res["OK"]:
        gLogger.error(res["Message"])
        Dexit(1)
    gLogger.notice(res["Value"])


if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.FrameworkSystem.private.standardLogging.LoggingRoot import LoggingRoot

gLogger = LoggingRoot()


def getLogger():
    return gLogger



""" Test_RSS_Policy_AlwaysActivePolicy
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id: $"

import unittest

import DIRAC.ResourceStatusSystem.Policy.CEAvailabilityPolicy as moduleTested


################################################################################


class CEAvailabilityPolicy_TestCase(unittest.TestCase):
    def setUp(self):
        """
        Setup
        """

        self.moduleTested = moduleTested
        self.testClass = self.moduleTested.CEAvailabilityPolicy

    def tearDown(self):
        """
        TearDown
        """
        del self.testClass
        del self.moduleTested


################################################################################
# Tests


class CEAvailabilityPolicy_Success(CEAvailabilityPolicy_TestCase):
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class"""

        policy = self.testClass()
        self.assertEqual("CEAvailabilityPolicy", policy.__class__.__name__)

    def test_evaluate(self):
        """tests the evaluate method"""

        policy = self.testClass()
        commandResult = {
            "OK": True,
            "Value": {
                "Reason": "All queues in 'Production'",
                "Status": "Production",
                "cccreamceli05.in2p3.fr:8443/cream-sge-long": "Production",
                "cccreamceli05.in2p3.fr:8443/cream-sge-verylong": "Production",
            },
        }
        res = policy._evaluate(commandResult)

        self.assertTrue(res["OK"])
        self.assertEqual("Active", res["Value"]["Status"])

        commandResult = {
            "OK": True,
            "Value": {
                "Reason": "All queues in 'Production'",
                "Status": "Degraded",
                "cccreamceli05.in2p3.fr:8443/cream-sge-long": "Production",
                "cccreamceli05.in2p3.fr:8443/cream-sge-verylong": "Production",
            },
        }
        res = policy._evaluate(commandResult)

        self.assertTrue(res["OK"])
        self.assertEqual("Banned", res["Value"]["Status"])


################################################################################

if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(CEAvailabilityPolicy_TestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(CEAvailabilityPolicy_Success))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



########################################################################
# $HeadURL $
# File: OperationHandlerBaseTests.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2013/03/25 08:09:08
########################################################################

""" :mod: OperationHandlerBaseTests
    ===============================

    .. module: OperationHandlerBaseTests
    :synopsis: unittests for OperationHandlerBase
    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    unittests for OperationHandlerBase
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

# #
# @file OperationHandlerBaseTests.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2013/03/25 08:09:21
# @brief Definition of OperationHandlerBaseTests class.

# # imports
import unittest
from DIRAC.RequestManagementSystem.private.OperationHandlerBase import OperationHandlerBase
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.DataManagementSystem.Client.DataManager import DataManager

########################################################################
class OperationHandlerBaseTests(unittest.TestCase):
    """
    .. class:: OperationHandlerBaseTests

    """

    def setUp(self):
        """test set up"""
        self.req = Request()
        self.req.RequestName = "testRequest"
        self.op = Operation({"Type": "ForwardDISET", "Arguments": "foobar"})
        self.req.addOperation(self.op)
        self.baseOp = OperationHandlerBase()

    def tearDown(self):
        """test tear down"""
        del self.baseOp
        del self.op
        del self.req

    def testOperationHandlerBase(self):
        """base op test"""
        self.baseOp.setOperation(self.op)

        # # log is there
        self.assertEqual("log" in dir(self.baseOp), True, "log missing")
        # # operation is there
        self.assertEqual("operation" in dir(self.baseOp), True, "operation is missing")
        # # request is there
        self.assertEqual("request" in dir(self.baseOp), True, "request is missing")
        # # __call__ not implemented
        self.assertRaises(NotImplementedError, self.baseOp)
        # # replica manager
        self.assertEqual(isinstance(self.baseOp.dm, DataManager), True, "DataManager is missing")


# # tests execution
if __name__ == "__main__":
    testLoader = unittest.TestLoader()
    OperationHandlerBaseTests = testLoader.loadTestsFromTestCase(OperationHandlerBaseTests)
    suite = unittest.TestSuite([OperationHandlerBaseTests])
    unittest.TextTestRunner(verbosity=3).run(suite)
